"""
"""


# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import logging

import numpy as np
import pandas as pd
import statsmodels.api as sm

from ..core import StatsModels, StatsError


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


__all__ = ["StatsMixedLM"]


logger = logging.getLogger(__name__)


class StatsMixedLM(StatsModels):
    def __init__(self, reml=True):
        """Initializes a 'StatsMixedLM' instance.

        Args:
            reml (bool): Whether to use REML or ML for the test.

        """
        self._reml = reml

    @staticmethod
    def _prepare_data(y, X):
        """Prepares the data for the MixedLM analysis."""
        # Checking y as the outcome and the groups
        missing_cols = {"outcome", "groups"} - set(y.columns)
        if missing_cols:
            raise ValueError(
                "missing column in y: mixedlm requires 'outcome' and 'groups'"
            )
        if len(y.columns) > 2:
            extra = set(y.columns) - {"outcome", "groups"}
            logger.warning("{}: unknown column in y, will be ignored".format(
                ",".join(extra),
            ))

        return y[["outcome"]], X, y.groups

    @staticmethod
    def _format_re(re):
        """Formats the random effects."""
        # Statsmodels returns a dictionary for the random effects
        return pd.Series(
            {sample: value.iloc[0] for sample, value in re.items()}
        )

    def fit(self, y, X):
        """Fit the model.

        Args:
            y (pandas.DataFrame): The vector of endogenous variable.
            X (pandas.DataFrame): The matrix of exogenous variables.

        """
        # Retrieving the data
        y, X, groups = self._prepare_data(y, X)

        # Creating the MixedLM model from StatsModels and fitting it
        model = sm.MixedLM(y, X, groups)
        try:
            # fitted = model.fit(reml=self._reml)
            fitted = model.fit(reml=self._reml)

        except np.linalg.linalg.LinAlgError as e:
            raise StatsError(str(e))

        out = {}
        parameters = fitted.params.index

        # Results about the model fit
        out = {
            "MODEL": {
                "log_likelihood": fitted.llf,
                "nobs": X.shape[0],
                "n_groups": model.n_groups,
                "random_effects": self._format_re(fitted.random_effects),
            },
        }

        # Getting the confidence intervals
        conf_ints = fitted.conf_int()

        for param in parameters:
            # If GWAS, check that inference could be done on the SNP
            if param == "SNPs" and np.isnan(fitted.pvalues[param]):
                raise StatsError("Inference did not converge.")

            out[param] = {
                "coef": fitted.params[param],
                "std_err": fitted.bse[param],
                "lower_ci": conf_ints.loc[param, 0],
                "upper_ci": conf_ints.loc[param, 1],
                "z_value": fitted.tvalues[param],
                "p_value": fitted.pvalues[param],
            }

        return out



"""
Small utility to convert the output from imputed-stats to genetest.
"""

# This file is part of genetest
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import sys


EXPECTED_COLUMNS = {
    "chr",
    "pos",
    "snp",
    "major",
    "minor",
    "maf",
    "n",
    "coef",
    "se",
    "lower",
    "upper",
    "t",
    "p",
}


def print_usage():
    print("Usage:")
    print("imputed-stats2genetest my_file.linear.dosage")
    quit(1)


def print_header(linear):
    cols = ["snp", "chr", "pos", "major", "minor", "maf", "n", "ll"]

    if linear:
        cols.append("adj_r2")

    cols += ["coef", "se", "lower", "upper", "t", "p"]

    print(*cols, sep="\t")


def convert_file(f):
    header = next(f).strip().split("\t")
    header = {col: i for i, col in enumerate(header)}

    linear = "adj.r-squared" in header

    # Check that the file has the appropriate columns.
    missing_cols = EXPECTED_COLUMNS - set(header.keys())
    if missing_cols:
        raise ValueError("Could not find column(s): {}.".format(missing_cols))

    print_header(linear)
    for line in f:
        line = line.strip().split("\t")
        cols = [
            line[header["snp"]],
            line[header["chr"]],
            line[header["pos"]],
            line[header["major"]],
            line[header["minor"]],
            line[header["maf"]],
            line[header["n"]],
            "NA",  # log-likelihood.
        ]

        if linear:
            cols.append(line[header["adj.r-squared"]])

        cols += [
            line[header["coef"]],
            line[header["se"]],
            line[header["lower"]],
            line[header["upper"]],
            line[header["t"]],
            line[header["p"]],
        ]

        print(*cols, sep="\t")


def main():
    if len(sys.argv) != 2:
        print_usage()

    filename = sys.argv[1]
    try:
        with open(filename, "r") as f:
            convert_file(f)
    except FileNotFoundError:
        print_usage()



"""
"""


# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


from .models.linear import StatsLinear
from .models.survival import StatsCoxPH
from .models.mixedlm import StatsMixedLM
from .models.logistic import StatsLogistic


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "CC BY-NC 4.0"


__all__ = ["available_models"]


available_models = {
    "linear": "Linear regression (ordinary least squares).",
    "logistic": "Logistic regression (GLM with binomial distribution).",
    "mixedlm": "Linear mixed effect model (random intercept).",
    "coxph": "Cox's proportional hazard regression (survival analysis).",
}


# The model map (which maps the name to the class)
model_map = {
    "linear": StatsLinear,
    "logistic": StatsLogistic,
    "mixedlm": StatsMixedLM,
    "coxph": StatsCoxPH,
}



"""
"""


# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


__all__ = ["PhenotypesContainer"]


class PhenotypesContainer(object):
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        raise NotImplementedError()

    def get_phenotypes(self, li=None):
        """Returns a dataframe of phenotypes.

        Args:
            li (list): A list of phenotypes to extract (default: None means
                       all phenotypes).

        Returns:
            pandas.DataFrame: A dataframe containing the phenotypes (with the
            sample IDs as index).

        """
        raise NotImplementedError()

    def get_nb_samples(self):
        """Returns the number of samples.

        Returns:
            int: The number of samples.

        """
        raise NotImplementedError()

    def get_nb_variables(self):
        """Returns the number of variables.

        Returns:
            int: The number of variables.

        """
        raise NotImplementedError()

    def is_repeated(self):
        """Check if the phenotypes contain repeated measurements.

        Returns:
            bool: ``True`` if the data contains repeated measurements,
            ``False`` otherwise.

        """
        raise NotImplementedError()

    def keep_samples(self, keep):
        """Keeps only a subset of samples.

        Args:
            keep (list): The list of samples to keep.

        """
        raise NotImplementedError()

    def get_sex(self):
        """Returns the sex for all samples.

        Returns:
            pandas.Series: The sex for all samples.

        """
        if not hasattr(self, "_sex_column") or self._sex_column is None:
            raise ValueError("No sex column was specified in Phenotype "
                             "container")

        # Extracting the sex
        sex = self._phenotypes.loc[:, self._sex_column]

        if self._repeated:
            # Checking if the sex are the same across repeated measures
            invalid = []
            group = sex.groupby(sex.index)
            for sample, values in group:
                if len(values.unique()) != 1:
                    invalid.append(sample)

            if len(invalid) > 0:
                raise ValueError(
                    "Samples with duplicated different sex: {}"
                    "".format(",".join(sorted(invalid))),
                )

            sex = group.first()

        # Checking we only have 0 and 1 (and maybe NaN)
        if len(set(sex.dropna().unique()) - {0.0, 1.0}) != 0:
            raise ValueError("Sex should only contain 0 and 1 (and maybe NaN)")

        return sex



"""
"""


# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


from .text import TextPhenotypes


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "CC BY-NC 4.0"


__all__ = ["available_formats", "format_map"]


available_formats = dict(
    text="Normal text file (TSV, CSV, etc).",
)


# The format map (which maps the name to the class)
format_map = dict(
    text=TextPhenotypes,
)



"""
"""


# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import numpy as np
import statsmodels.api as sm

from ..core import StatsModels, StatsError


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


__all__ = ["StatsLinear"]


class StatsLinear(StatsModels):
    def __init__(self, condition_value_t=1000, eigenvals_t=1e-10):
        """Initializes a 'StatsLinear' instance.

        Args:
            condition_value_t (int): The condition value threshold (for
                                     multicollinearity). Usually, values higher
                                     than 1000 indicate strong
                                     multicollinearity or other numerical
                                     problems.
            eigenvals_t (float): The Eigen value threshold (for
                                 multicollinearity). Usually, values lower than
                                 1e-10 might indicate strong multicollinearity
                                 or singular design matrix.

        """
        # Saving the condition value threshold
        self._condition_value_t = condition_value_t
        self._eigenvals_t = eigenvals_t

    def _results_handler(self, fitted):
        # Checking the condition number (according to StatsModels, condition
        # number higher than 1000 indicate that there are strong
        # multicollinearity or other numerical problems)
        if fitted.condition_number > self._condition_value_t:
            raise StatsError("condition number is large, {}".format(
                fitted.condition_number,
            ))

        # Checking the smallest eigenvalue (according to StatsModels, values
        # lower than 1e-10 might indicate that there are strong
        # multicollinearity problems or that the design matrix is singular)
        if fitted.eigenvals.min() < self._eigenvals_t:
            raise StatsError("smallest eigenvalue is small, {}".format(
                fitted.eigenvals.min(),
            ))

        # Results about the model fit.
        out = {
            "MODEL": {
                "r_squared_adj": fitted.rsquared_adj,
                "log_likelihood": fitted.llf,
                "nobs": fitted.nobs
            }
        }

        # Getting the confidence intervals
        conf_ints = fitted.conf_int()

        # Results about individual model parameters.
        parameters = fitted.params.index
        for param in parameters:
            # If GWAS, check that inference could be done on the SNP.
            if param == "SNPs" and np.isnan(fitted.pvalues[param]):
                raise StatsError("Inference did not converge.")

            out[param] = {
                "coef": fitted.params[param],
                "std_err": fitted.bse[param],
                "lower_ci": conf_ints.loc[param, 0],
                "upper_ci": conf_ints.loc[param, 1],
                "t_value": fitted.tvalues[param],
                "p_value": fitted.pvalues[param],
            }

        return out

    def fit(self, y, X, handler=None):
        """Fit the model.

        Args:
            y (pandas.DataFrame): The vector of endogenous variable.
            X (pandas.DataFrame): The matrix of exogenous variables.

        """
        # Creating the OLS model from StatsModels and fitting it
        model = sm.OLS(y, X)
        handler = self._results_handler if handler is None else handler
        return handler(model.fit())





# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import os
import unittest
from tempfile import TemporaryDirectory

import numpy as np
import pandas as pd

from ..phenotypes.core import PhenotypesContainer
from ..phenotypes.text import TextPhenotypes


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


class TestCore(unittest.TestCase):
    def test_get_phenotypes(self):
        """Tests that 'get_phenotypes' raises a NotImplementedError."""
        with self.assertRaises(NotImplementedError):
            PhenotypesContainer().get_phenotypes()

    def test_close(self):
        """Tests that 'close' raises a NotImplementedError."""
        with self.assertRaises(NotImplementedError):
            PhenotypesContainer().close()

    def test_get_sex_no_attribute(self):
        """Tests when the '_sex_column' attribute is missing."""
        with self.assertRaises(ValueError):
            PhenotypesContainer().get_sex()


class TestText(unittest.TestCase):
    def setUp(self):
        self.tmpdir = TemporaryDirectory(prefix="genetest_")

        # The phenotype file
        pheno_file = os.path.join(self.tmpdir.name, "pheno.txt")
        with open(pheno_file, "w") as f:
            print("sample", "V1", "V2", "V3", "sex", file=f)
            print("s1", "1", "1.2", "a", "0", file=f)
            print("s2", "234", "34.3", "sdc", "1", file=f)
            print("s3", "-999999", "", "csgb", "1", file=f)
            print("s4", "5463", "16.8", "999999", "0", file=f)
            print("s5", "2356", "574.8", "jyrhg", "0", file=f)
            print("s6", "", "567.2", "bvnchy", "1", file=f)
            print("s7", "57634", "3134.3", "", "0", file=f)
            print("s8", "346", "999999", "dfgfjgsd", "", file=f)
            print("s9", "34626", "3421.3", "fhyrj", "0", file=f)

        # The expected data frame
        self.expected_pheno = pd.DataFrame(
            [("s1", 1, 1.2, "a", 0),
             ("s2", 234, 34.3, "sdc", 1),
             ("s3", -999999, np.nan, "csgb", 1),
             ("s4", 5463, 16.8, "999999", 0),
             ("s5", 2356, 574.8, "jyrhg", 0),
             ("s6", np.nan, 567.2, "bvnchy", 1),
             ("s7", 57634, 3134.3, np.nan, 0),
             ("s8", 346, 999999, "dfgfjgsd", np.nan),
             ("s9", 34626, 3421.3, "fhyrj", 0)],
            columns=["sample", "V1", "V2", "V3", "sex"],
        ).set_index("sample")

        # The TextPhenotypes parameters
        self.parameters = dict(
            filename=pheno_file,
            sample_column="sample",
            field_separator=" ",
            missing_values=None,
            repeated_measurements=False,
            sex_column="sex",
        )

    def tearDown(self):
        self.tmpdir.cleanup()

    def test_normal_functionality(self):
        """Tests normal functionality."""
        observed_pheno = TextPhenotypes(**self.parameters)

        self.assertTrue(
            self.expected_pheno.equals(observed_pheno.get_phenotypes())
        )
        self.assertEqual(9, observed_pheno.get_nb_samples())
        self.assertFalse(observed_pheno.is_repeated())

    def test_subset_phenotypes(self):
        """Tests asking for a subset of phenotypes."""
        container = TextPhenotypes(**self.parameters)

        subset = ("V1", "V3")
        expected = self.expected_pheno.loc[:, ["V1", "V3"]]
        self.assertTrue(
            expected.equals(container.get_phenotypes(subset))
        )

    def test_bad_subset_phenotypes(self):
        """Tests asking for a subset with unavailable phenotypes."""
        container = TextPhenotypes(**self.parameters)
        with self.assertRaises(KeyError):
            container.get_phenotypes(("V1", "V3", "VX"))

    def test_get_sex(self):
        """Tests asking for the sex series."""
        container = TextPhenotypes(**self.parameters)

        expected = self.expected_pheno.loc[:, "sex"]
        self.assertTrue(
            expected.equals(container.get_sex()),
        )

    def test_repeated_measurements(self):
        """Tests when there are repeated measurements available."""
        with open(self.parameters["filename"], "w") as f:
            print("sample", "Time", "V1", "V2", "V3", "sex", file=f)
            print("s1", "13", "1", "1.2", "a", "0", file=f)
            print("s1", "125", "1", "1.2", "a", "0", file=f)
            print("s1", "356", "1", "1.2", "a", "0", file=f)
            print("s2", "12", "5463", "16.8", "b", "1", file=f)
            print("s2", "34", "5463", "16.8", "b", "1", file=f)
            print("s2", "67", "5463", "16.8", "b", "1", file=f)
            print("s3", "1", "57634", "3134.3", "c", "", file=f)
            print("s3", "5", "57634", "3134.3", "c", "", file=f)
            print("s3", "10", "57634", "3134.3", "c", "", file=f)
            print("s4", "2", "57635", "3134.6", "b", "1", file=f)
            print("s4", "6", "57635", "3134.6", "b", "1", file=f)
            print("s4", "11", "57635", "3134.6", "b", "1", file=f)

        expected_pheno = pd.DataFrame(
            [("s1", 13, 1, 1.2, "a", 0),
             ("s1", 125, 1, 1.2, "a", 0),
             ("s1", 356, 1, 1.2, "a", 0),
             ("s2", 12, 5463, 16.8, "b", 1),
             ("s2", 34, 5463, 16.8, "b", 1),
             ("s2", 67, 5463, 16.8, "b", 1),
             ("s3", 1, 57634, 3134.3, "c", np.nan),
             ("s3", 5, 57634, 3134.3, "c", np.nan),
             ("s3", 10, 57634, 3134.3, "c", np.nan),
             ("s4", 2, 57635, 3134.6, "b", 1),
             ("s4", 6, 57635, 3134.6, "b", 1),
             ("s4", 11, 57635, 3134.6, "b", 1)],
            columns=["sample", "Time", "V1", "V2", "V3", "sex"],
        ).set_index("sample")

        # Changing the parameters
        self.parameters["repeated_measurements"] = True

        # Getting the observed phenotypes
        observed_pheno = TextPhenotypes(**self.parameters)

        # Checking the values
        self.assertTrue(expected_pheno.equals(observed_pheno.get_phenotypes()))
        self.assertEqual(4, observed_pheno.get_nb_samples())
        self.assertTrue(observed_pheno.is_repeated())

        # Checking the sex specifically
        expected_sex = pd.Series(
            [0, 1, np.nan, 1], index=["s1", "s2", "s3", "s4"],
        )
        self.assertTrue(expected_sex.equals(observed_pheno.get_sex()))

    def test_one_other_missing_value(self):
        """Tests when there are other missing values."""
        # Changing the parameters
        self.parameters["missing_values"] = "999999"

        # Getting the observed phenotypes
        observed_pheno = TextPhenotypes(**self.parameters).get_phenotypes()

        # Changing the expected phenotypes
        self.expected_pheno.loc["s4", "V3"] = np.nan
        self.expected_pheno.loc["s8", "V2"] = np.nan

        # Comparing
        self.assertTrue(self.expected_pheno.equals(observed_pheno))

    def test_multiple_missing_values(self):
        """Tests when there are multiple missing values."""
        # Changing the parameters
        self.parameters["missing_values"] = ["-999999", "999999"]

        # Getting the observed phenotypes
        observed_pheno = TextPhenotypes(**self.parameters).get_phenotypes()

        # Changing the expected phenotypes
        self.expected_pheno.loc["s3", "V1"] = np.nan
        self.expected_pheno.loc["s4", "V3"] = np.nan
        self.expected_pheno.loc["s8", "V2"] = np.nan

        # Comparing
        self.assertTrue(self.expected_pheno.equals(observed_pheno))

    def test_specific_column_missing_value(self):
        """Tests when there are specific missing values for some columns."""
        # Changing the parameters
        self.parameters["missing_values"] = {"V3": ["999999"]}

        # Getting the observed phenotypes
        observed_pheno = TextPhenotypes(**self.parameters).get_phenotypes()

        # Changing the expected phenotypes
        self.expected_pheno.loc["s4", "V3"] = np.nan

        # Comparing
        self.assertTrue(self.expected_pheno.equals(observed_pheno))

    def test_string_representation(self):
        """Checks the string representation."""
        with TextPhenotypes(**self.parameters) as text_pheno:
            self.assertEqual(
                "TextPhenotypes(9 samples, 4 variables)",
                str(text_pheno),
            )

    def test_invalid_sex(self):
        """Tests when other values than 0, 1 and NaN are in the sex column."""
        with open(self.parameters["filename"], "w") as f:
            print("sample", "V1", "V2", "V3", "sex", file=f)
            print("s1", "1", "1.2", "a", "0", file=f)
            print("s2", "234", "34.3", "sdc", "1", file=f)
            print("s3", "-999999", "", "csgb", "1", file=f)
            print("s4", "5463", "16.8", "999999", "2", file=f)
            print("s5", "2356", "574.8", "jyrhg", "0", file=f)
            print("s6", "", "567.2", "bvnchy", "1", file=f)
            print("s7", "57634", "3134.3", "", "0", file=f)
            print("s8", "346", "999999", "dfgfjgsd", "", file=f)
            print("s9", "34626", "3421.3", "fhyrj", "0", file=f)

        container = TextPhenotypes(**self.parameters)
        with self.assertRaises(ValueError):
            container.get_sex()

    def test_undefined_sex_column(self):
        """Tests when the sex column wasn't specified."""
        del self.parameters["sex_column"]

        container = TextPhenotypes(**self.parameters)
        with self.assertRaises(ValueError):
            container.get_sex()

    def test_duplicated_sex(self):
        """Tests when there are repeated (different) sex for a sample."""
        with open(self.parameters["filename"], "w") as f:
            print("sample", "Time", "V1", "V2", "V3", "sex", file=f)
            print("s1", "13", "1", "1.2", "a", "0", file=f)
            print("s1", "125", "1", "1.2", "a", "0", file=f)
            print("s1", "356", "1", "1.2", "a", "1", file=f)
            print("s2", "12", "5463", "16.8", "b", "1", file=f)
            print("s2", "34", "5463", "16.8", "b", "1", file=f)
            print("s2", "67", "5463", "16.8", "b", "", file=f)
            print("s3", "1", "57634", "3134.3", "c", "", file=f)
            print("s3", "5", "57634", "3134.3", "c", "", file=f)
            print("s3", "10", "57634", "3134.3", "c", "", file=f)

        # Changing the parameters
        self.parameters["repeated_measurements"] = True

        # Getting the observed phenotypes
        container = TextPhenotypes(**self.parameters)
        with self.assertRaises(ValueError) as cm:
            container.get_sex()
        self.assertEqual(
            str(cm.exception), "Samples with duplicated different sex: s1,s2",
        )





# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import unittest

import numpy as np
import pandas as pd

from ..statistics.core import StatsModels


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


@unittest.skip("Not implemented")
class TestStatsLinear(unittest.TestCase):
    """Tests the 'StatsLinear' class."""
    def setUp(self):
        # The data
        data = pd.DataFrame(
            [("s0", 3.0, 12.0, "f1"), ("s1", 3.4, 30.8, "f2"),
             ("s2", 5.3, 50.2, "f1"), ("s3", 6.0, 30.6, "f3"),
             ("s4", 0.5, 40.0, "f2"), ("s5", 2.4, 80.5, "f1"),
             ("s6", 5.6, 70.0, "f1"), ("s7", 7.6, 87.4, "f2"),
             ("s8", 0.3, 63.0, "f1"), ("s9", 1.9, 54.3, "f3")],
            columns=["sample_id", "pheno", "var1", "var2"],
        ).set_index("sample_id")
        self.data = data.iloc[np.random.permutation(len(data))]

        # The genotypes
        genotypes = pd.DataFrame(
            [("s0", 0.0), ("s1", 1.0), ("s2", 0.0), ("s3", np.nan),
             ("s4", 0.0), ("s5", 1.0), ("s6", 2.0), ("s7", np.nan),
             ("s8", 0.0), ("s9", 0.0)],
            columns=["iid", "geno"],
        ).set_index("iid")
        self.genotypes = genotypes.iloc[np.random.permutation(len(genotypes))]

        # The expected values for y and X
        self.expected_y = pd.DataFrame(
            [("s0", 3.0), ("s1", 3.4), ("s2", 5.3), ("s3", 6.0), ("s4", 0.5),
             ("s5", 2.4), ("s6", 5.6), ("s7", 7.6), ("s8", 0.3), ("s9", 1.9)],
            columns=["sample_id", "pheno"],
        ).set_index("sample_id")
        self.expected_X = pd.DataFrame(
            [("s0", 1.0, 0.0, 0.0, 12.0), ("s1", 1.0, 1.0, 0.0, 30.8),
             ("s2", 1.0, 0.0, 0.0, 50.2), ("s3", 1.0, 0.0, 1.0, 30.6),
             ("s4", 1.0, 1.0, 0.0, 40.0), ("s5", 1.0, 0.0, 0.0, 80.5),
             ("s6", 1.0, 0.0, 0.0, 70.0), ("s7", 1.0, 1.0, 0.0, 87.4),
             ("s8", 1.0, 0.0, 0.0, 63.0), ("s9", 1.0, 0.0, 1.0, 54.3)],
            columns=["sample_id", "Intercept", "C(var2)[T.f2]",
                     "C(var2)[T.f3]", "var1"],
        ).set_index("sample_id")

        # The expected values for the new y and the new X
        self.expected_new_y = pd.DataFrame(
            [("s0", 3.0), ("s1", 3.4), ("s2", 5.3), ("s4", 0.5), ("s5", 2.4),
             ("s6", 5.6), ("s8", 0.3), ("s9", 1.9)],
            columns=["sample_id", "pheno"],
        ).set_index("sample_id")
        self.expected_new_X = pd.DataFrame(
            [("s0", 1.0, 0.0, 0.0, 12.0, 0.0),
             ("s1", 1.0, 1.0, 0.0, 30.8, 1.0),
             ("s2", 1.0, 0.0, 0.0, 50.2, 0.0),
             ("s4", 1.0, 1.0, 0.0, 40.0, 0.0),
             ("s5", 1.0, 0.0, 0.0, 80.5, 1.0),
             ("s6", 1.0, 0.0, 0.0, 70.0, 2.0),
             ("s8", 1.0, 0.0, 0.0, 63.0, 0.0),
             ("s9", 1.0, 0.0, 1.0, 54.3, 0.0)],
            columns=["sample_id", "Intercept", "C(var2)[T.f2]",
                     "C(var2)[T.f3]", "var1", "geno"],
        ).set_index("sample_id")

        # Creating an instance of StatsModels core object
        self.core_model = StatsModels(
            outcomes=["test"],
            predictors=["var1", "var2", "var3"],
            interaction=None,
            intercept=True,
        )

    def test__create_model(self):
        """Tests the '_create_model' function."""
        self.core_model._create_model(
            outcomes=["y"],
            predictors=["var1", "C(var2)"],
            interaction=None,
            intercept=True,
        )

        # Checking the model (by looking at the formula)
        self.assertEqual(
            "y ~ geno + var1 + C(var2)",
            self.core_model.get_model_description(),
        )

        # Checking the resulting column
        self.assertEqual("geno", self.core_model._result_col)

    def test__create_model_no_intercept(self):
        """Tests the '_create_model' function with no intercept."""
        self.core_model._create_model(
            outcomes=["y"],
            predictors=["var1", "C(var2)"],
            interaction=None,
            intercept=False,
        )

        # Checking the model (by looking at the formula)
        self.assertEqual(
            "y ~ 0 + geno + var1 + C(var2)",
            self.core_model.get_model_description(),
        )

        # Checking the resulting column
        self.assertEqual("geno", self.core_model._result_col)

    def test__create_model_with_interaction(self):
        """Tests the '_create_model' function with an interaction."""
        self.core_model._create_model(
            outcomes=["y"],
            predictors=["var1", "C(var2)"],
            interaction="var1",
            intercept=False,
        )

        # Checking the model (by looking at the formula)
        self.assertEqual(
            "y ~ 0 + geno + var1 + C(var2) + geno:var1",
            self.core_model.get_model_description(),
        )

        # Checking the resulting column
        self.assertEqual("geno:var1", self.core_model._result_col)

    def test__create_model_with_interaction2(self):
        """Tests with an interaction (categorical)."""
        self.core_model._create_model(
            outcomes=["y"],
            predictors=["var1", "C(var2)"],
            interaction="C(var2)",
            intercept=True,
        )

        # Checking the model (by looking at the formula)
        self.assertEqual(
            "y ~ geno + var1 + C(var2) + geno:C(var2)",
            self.core_model.get_model_description(),
        )

        # Checking the resulting column
        self.assertEqual("geno:C(var2)", self.core_model._result_col)

    def test__create_model_with_interaction_not_in_predictors(self):
        """Tests with an interaction but not in predictors list."""
        self.core_model._create_model(
            outcomes=["y"],
            predictors=["var1", "C(var2)"],
            interaction="var3",
            intercept=True,
        )

        # Checking the model (by looking at the formula)
        self.assertEqual(
            "y ~ geno + var1 + C(var2) + var3 + geno:var3",
            self.core_model.get_model_description(),
        )

        # Checking the resulting column
        self.assertEqual("geno:var3", self.core_model._result_col)





# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import unittest
from os import path
from tempfile import TemporaryDirectory

import numpy as np
import pandas as pd
from scipy.stats import binom

from pyplink import PyPlink
from geneparse import parsers

from .. import modelspec as spec
from ..phenotypes.dummy import _DummyPhenotypes


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


class TestModelSpec(unittest.TestCase):
    """Tests the 'ModelSpec' class."""
    @classmethod
    def setUpClass(cls):
        # Creating random data
        cls.data = pd.DataFrame(
            dict(
                pheno=np.random.randint(1, 100, 100),
                var1=np.random.randint(1, 100, 100),
                var2=np.random.rand(100),
                var3=["x{}".format(i) for i in np.random.randint(0, 3, 100)],
                var4=["y{}".format(i) for i in np.random.randint(0, 2, 100)],
                var5=np.random.randint(0, 4, 100),
                snp=binom.rvs(2, 0.3, size=100),
            ),
            index=["sample_{}".format(i+1) for i in range(100)],
        )

        # Changing one factor to categorical data
        cls.data.loc[:, "var5"] = cls.data.var5.astype("category")

        # Creating the dummy phenotype container
        phenotypes = ["pheno"] + ["var{}".format(i+1) for i in range(5)]
        cls.phenotypes = _DummyPhenotypes()
        cls.phenotypes.data = cls.data[phenotypes].copy()

        # Creating a temporary directory
        cls.tmp_dir = TemporaryDirectory(prefix="genetest_")

        # The plink file prefix
        cls.plink_prefix = path.join(cls.tmp_dir.name, "input")

        # Permuting the sample to add a bit of randomness
        new_sample_order = np.random.permutation(
            np.copy(cls.data.index.values)
        )

        # Creating the BED file
        with PyPlink(cls.plink_prefix, "w") as bed:
            bed.write_genotypes(cls.data.loc[new_sample_order, "snp"])

        # Creating the BIM file
        with open(cls.plink_prefix + ".bim", "w") as bim:
            print(1, "snp", 0, 1, "B", "A", sep="\t", file=bim)

        # Creating the FAM file
        with open(cls.plink_prefix + ".fam", "w") as fam:
            for sample in new_sample_order:
                print(sample, sample, 0, 0, 0, -9, file=fam)

        # Creating the genotype parser
        cls.genotypes = parsers["plink"](cls.plink_prefix)

    @classmethod
    def tearDownClass(cls):
        cls.genotypes.close()
        cls.tmp_dir.cleanup()

    def setUp(self):
        # Resetting the model specification
        spec._reset()

        # Reordering the columns and the rows of the phenotype data frame
        self.phenotypes.data = self.phenotypes.data.iloc[
            np.random.permutation(self.phenotypes.data.shape[0]),
            np.random.permutation(self.phenotypes.data.shape[1])
        ]

    def test_reset(self):
        """Tests the reset function of the ModelSpec."""
        # Checking there are no transformations
        transformations = spec.core.TransformationManager.transformations
        self.assertEqual(0, len(transformations),
                         "Transformations were not reset")

        # Checking there are no dependencies
        dependencies = spec.core.DependencyManager.dependencies
        self.assertEqual(0, len(dependencies),
                         "Dependencies were not reset")

    def test_simple_modelspec(self):
        """Tests a simple ModelSpec object."""
        # Creating the model specification
        predictors = [spec.genotypes.snp, spec.phenotypes.var1,
                      spec.phenotypes.var2]
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno,
            predictors=predictors,
            test="linear",
        )

        # Gathering the observed matrix
        matrix = modelspec.create_data_matrix(
            self.phenotypes, self.genotypes,
        )

        # Checking the shape of the matrix
        self.assertEqual((self.data.shape[0], 5), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[self.data.index, outcome_col]
        self.assertTrue(outcomes.equals(self.data.pheno),
                        "The outcomes are not as expected")

        # Checking the predictors
        translations = modelspec.get_translations()
        for predictor in predictors:
            # Getting the name of the predictor
            name = translations[predictor.id]

            # Comparing the values
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, predictor.id].values,
                self.data[name].values,
                err_msg="The predictor '{}' is not as expected".format(name),
            )

    def test_no_intersect(self):
        """Tests when no intersect between phenotypes and genotypes."""
        # Creating a new phenotype container
        phenotypes = _DummyPhenotypes()
        phenotypes.data = self.data[["pheno", "var1"]].copy()
        phenotypes.data.index = ["s" + s for s in phenotypes.data.index]

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno,
            predictors=[spec.genotypes.snp, spec.phenotypes.var1],
            test="linear",
        )

        # Gathering the observed matrix
        with self.assertRaises(ValueError):
            modelspec.create_data_matrix(phenotypes, self.genotypes)

    def test_smaller_intersect(self):
        """Tests when the sample intersect is smaller between containers."""
        # Choosing 10 samples to exclude from the dataset
        to_exclude = np.random.choice(self.data.index, 10, replace=False)

        # Removing 5 samples from the phenotypes
        phenotypes = _DummyPhenotypes()
        phenotypes.data = self.data.drop(to_exclude[:5], axis=0)

        # Removing the next 5 for the genotypes
        plink_prefix = self.plink_prefix + "_less"
        geno_data = self.data.drop(to_exclude[5:], axis=0)
        with PyPlink(plink_prefix, "w") as bed:
            bed.write_genotypes(geno_data.snp)

        # Creating the BIM file
        with open(plink_prefix + ".bim", "w") as bim:
            print(1, "snp", 0, 1, "B", "A", sep="\t", file=bim)

        # Creating the FAM file
        with open(plink_prefix + ".fam", "w") as fam:
            for sample in geno_data.index:
                print(sample, sample, 0, 0, 0, -9, file=fam)

        # Creating the model specification
        predictors = [spec.genotypes.snp, spec.phenotypes.var1,
                      spec.phenotypes.var2]
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno,
            predictors=predictors,
            test="linear",
        )

        # Gathering the observed matrix
        with parsers["plink"](plink_prefix) as genotypes:
            matrix = modelspec.create_data_matrix(phenotypes, genotypes)

        # Subset of the new data
        data = self.data.drop(to_exclude, axis=0)

        # Checking the shape of the matrix
        self.assertEqual((data.shape[0], 5), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[data.index, outcome_col]
        self.assertTrue(outcomes.equals(data.pheno),
                        "The outcomes are not as expected")

        # Checking the predictors
        translations = modelspec.get_translations()
        for predictor in predictors:
            # Getting the name of the predictor
            name = translations[predictor.id]

            # Comparing the values
            np.testing.assert_array_equal(
                matrix.loc[data.index, predictor.id].values,
                data[name].values,
                err_msg="The predictor '{}' is not as expected".format(name),
            )

    def test_pow(self):
        """Tests a power transformation."""
        # Creating the model specification
        predictors = [spec.genotypes.snp, spec.pow(spec.phenotypes.var1, 3)]
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno,
            predictors=predictors,
            test="linear",
        )

        # Gathering the observed matrix
        matrix = modelspec.create_data_matrix(
            self.phenotypes, self.genotypes,
        )

        # Checking the shape of the matrix
        self.assertEqual((self.data.shape[0], 4), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[self.data.index, outcome_col]
        self.assertTrue(outcomes.equals(self.data.pheno),
                        "The outcomes are not as expected")

        # Checking the predictors
        for predictor, power, name in zip(predictors, (1, 3), ("snp", "var1")):
            # Comparing the values
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, predictor.id].values,
                self.data[name].values**power,
                err_msg="The predictor '{}**{}' is not as "
                        "expected".format(name, power),
            )

    def test_log(self):
        """Tests log transformations (log10 and ln)."""
        # Creating the model specification
        predictors = [spec.genotypes.snp, spec.log10(spec.phenotypes.var1),
                      spec.ln(spec.phenotypes.var1)]
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno,
            predictors=predictors,
            test="linear",
        )

        # Gathering the observed matrix
        matrix = modelspec.create_data_matrix(
            self.phenotypes, self.genotypes,
        )

        # Checking the shape of the matrix
        self.assertEqual((self.data.shape[0], 5), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[self.data.index, outcome_col]
        self.assertTrue(outcomes.equals(self.data.pheno),
                        "The outcomes are not as expected")

        # Checking the predictors
        translations = modelspec.get_translations()
        for predictor in predictors[:1]:
            # Getting the name of the predictor
            name = translations[predictor.id]

            # Comparing the values
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, predictor.id].values,
                self.data[name].values,
                err_msg="The predictor '{}' is not as expected".format(name),
            )

        # Checking the log transform
        for predictor, transform, name in zip(predictors[1:],
                                              (np.log10, np.log),
                                              ("var1", "var1")):
            # Comparing the values
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, predictor.id].values,
                transform(self.data[name].values),
                err_msg="The predictor '{}**{}' is not as "
                        "expected".format(name, transform.__name__),
            )

    def test_encode_factor(self):
        """Tests with factor."""
        # Creating the model specification
        predictors = [spec.factor(spec.phenotypes.var3),
                      spec.factor(spec.phenotypes.var4),
                      spec.factor(spec.phenotypes.var5)]
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno,
            predictors=predictors,
            test="linear",
        )

        # Gathering the observed matrix
        matrix = modelspec.create_data_matrix(
            self.phenotypes, self.genotypes,
        )

        # Checking the shape of the matrix
        self.assertEqual((self.data.shape[0], 8), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[self.data.index, outcome_col]
        self.assertTrue(outcomes.equals(self.data.pheno),
                        "The outcomes are not as expected")

        # Checking the factors
        predictor_zipped = zip(
            predictors,
            ("var3", "var4", "var5"),
            (("x1", "x2"), ("y1", ), (1, 2, 3)),
        )
        for predictor, name, levels in predictor_zipped:
            for level in levels:
                # Getting the name of the column containing the level data
                matrix_col = "{}:level.{}".format(predictor.id, level)

                # Comparing the results
                np.testing.assert_array_equal(
                    matrix.loc[self.data.index, matrix_col].values,
                    (self.data[name] == level).astype(float).values,
                    err_msg="The predictor '{}' (level '{}') is not as "
                            "expected".format(name, level),
                )

    def test_simple_interaction(self):
        """Tests simple interaction."""
        # Creating the model specification
        predictors = [spec.genotypes.snp, spec.phenotypes.var1]
        interaction = spec.interaction(spec.genotypes.snp,
                                       spec.phenotypes.var1)
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno,
            predictors=predictors + [interaction],
            test="linear",
        )

        # Gathering the observed matrix
        matrix = modelspec.create_data_matrix(
            self.phenotypes, self.genotypes,
        )

        # Checking the shape of the matrix
        self.assertEqual((self.data.shape[0], 5), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[self.data.index, outcome_col]
        self.assertTrue(outcomes.equals(self.data.pheno),
                        "The outcomes are not as expected")

        # Checking the predictors
        for predictor in predictors:
            # Getting the name of the predictor
            name = modelspec.get_translations()[predictor.id]

            # Comparing the values
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, predictor.id].values,
                self.data[name].values,
                err_msg="The predictor '{}' is not as expected".format(name),
            )

        # Checking the interaction
        np.testing.assert_array_equal(
            matrix.loc[self.data.index, interaction.id].values,
            (self.data.snp * self.data.var1).values,
            err_msg="The interaction 'snp*var1' is not as expected",
        )

    def test_multiple_interaction(self):
        """Tests simple interaction."""
        # Creating the model specification
        predictors = [spec.genotypes.snp, spec.phenotypes.var1,
                      spec.phenotypes.var2]
        interaction = spec.interaction(spec.genotypes.snp,
                                       spec.phenotypes.var1,
                                       spec.phenotypes.var2)
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno,
            predictors=predictors + [interaction],
            test="linear",
        )

        # Gathering the observed matrix
        matrix = modelspec.create_data_matrix(
            self.phenotypes, self.genotypes,
        )

        # Checking the shape of the matrix
        self.assertEqual((self.data.shape[0], 6), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[self.data.index, outcome_col]
        self.assertTrue(outcomes.equals(self.data.pheno),
                        "The outcomes are not as expected")

        # Checking the predictors
        for predictor in predictors:
            # Getting the name of the predictor
            name = modelspec.get_translations()[predictor.id]

            # Comparing the values
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, predictor.id].values,
                self.data[name].values,
                err_msg="The predictor '{}' is not as expected".format(name),
            )

        # Checking the interaction
        np.testing.assert_array_equal(
            matrix.loc[self.data.index, interaction.id].values,
            (self.data.snp * self.data.var1 * self.data.var2).values,
            err_msg="The interaction 'snp*var1*var2' is not as expected",
        )

    def test_factor_interaction(self):
        """Tests interaction between a term and a factor."""
        # Creating the model specification
        factor = spec.factor(spec.phenotypes.var5)
        predictors = [spec.genotypes.snp, factor]
        interaction = spec.interaction(spec.genotypes.snp, factor)
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno,
            predictors=predictors + [interaction],
            test="linear",
        )

        # Gathering the observed matrix
        matrix = modelspec.create_data_matrix(
            self.phenotypes, self.genotypes,
        )

        # Checking the shape of the matrix
        self.assertEqual((self.data.shape[0], 9), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[self.data.index, outcome_col]
        self.assertTrue(outcomes.equals(self.data.pheno),
                        "The outcomes are not as expected")

        # Checking the predictor
        np.testing.assert_array_equal(
            matrix.loc[self.data.index, spec.genotypes.snp.id].values,
            self.data.snp.values,
            err_msg="The predictor 'snp' is not as expected",
        )

        # For all level:
        for level in (1, 2, 3):
            # Creating the expected level
            expected_factor_values = (self.data.var5 == level).astype(float)

            # Checking the factor
            col = factor.id + ":level.{}".format(level)
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, col].values,
                expected_factor_values.values,
                err_msg="The predictor 'var5' (level '{}') is not as "
                        "expected".format(level),
            )

            # Checking the interaction
            col = interaction.id + ":level.{}".format(level)
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, col].values,
                (expected_factor_values * self.data.snp).values,
                err_msg="The interaction 'snp*var5' (var5 level '{}') is not "
                        "as expected".format(level),
            )

    def test_complex_interaction(self):
        """Tests a complex interaction between terms and factors."""
        # Creating the model specification
        var3 = spec.factor(spec.phenotypes.var3)
        var4 = spec.factor(spec.phenotypes.var4)
        var5 = spec.factor(spec.phenotypes.var5)
        predictors = [spec.genotypes.snp, spec.phenotypes.var2, var3, var4,
                      var5]
        interaction = spec.interaction(spec.genotypes.snp,
                                       spec.phenotypes.var2, var3, var4, var5)
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno,
            predictors=predictors + [interaction],
            test="linear",
        )

        # Gathering the observed matrix
        matrix = modelspec.create_data_matrix(
            self.phenotypes, self.genotypes,
        )

        # Checking the shape of the matrix
        self.assertEqual((self.data.shape[0], 16), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[self.data.index, outcome_col]
        self.assertTrue(outcomes.equals(self.data.pheno),
                        "The outcomes are not as expected")

        # Checking the predictors
        translations = modelspec.get_translations()
        for predictor in predictors[:2]:
            # Getting the name of the predictor
            name = translations[predictor.id]

            # Comparing the values
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, predictor.id].values,
                self.data[name].values,
                err_msg="The predictor '{}' is not as expected".format(name),
            )

        # Checking the factors
        predictor_zipped = zip(
            predictors[2:],
            ("var3", "var4", "var5"),
            (("x1", "x2"), ("y1", ), (1, 2, 3)),
        )
        for predictor, name, levels in predictor_zipped:
            for level in levels:
                # Getting the name of the column containing the level data
                matrix_col = "{}:level.{}".format(predictor.id, level)

                # Comparing the factor
                np.testing.assert_array_equal(
                    matrix.loc[self.data.index, matrix_col].values,
                    (self.data[name] == level).astype(float).values,
                    err_msg="The predictor '{}' (level '{}') is not as "
                            "expected".format(name, level),
                )

        # Checking the interaction
        for var3_l in ("x1", "x2"):
            for var4_l in ("y1", ):
                for var5_l in (1, 2, 3):
                    # The expected factor values
                    expected_var3 = (self.data.var3 == var3_l).astype(float)
                    expected_var4 = (self.data.var4 == var4_l).astype(float)
                    expected_var5 = (self.data.var5 == var5_l).astype(float)

                    # The expected interaction
                    expected_interaction = (
                        self.data.snp * self.data.var2 * expected_var3 *
                        expected_var4 * expected_var5
                    )

                    # Checking the interaction
                    matrix_col = "{}:level.{}:level.{}:level.{}".format(
                        interaction.id, var3_l, var4_l, var5_l,
                    )
                    np.testing.assert_array_equal(
                        matrix.loc[self.data.index, matrix_col].values,
                        expected_interaction.values,
                        err_msg=(
                            "The interaction 'snp*var2*var3*var4*var5*' "
                            "(var3 level '{}', var4 level '{} and var5 "
                            "level '{}' is not as expected".format(
                                var3_l, var4_l, var5_l,
                            )
                        ),
                    )

    def test_gwas_interaction(self):
        """Test a simple GWAS interaction."""
        # Creating the model specification
        inter = spec.gwas_interaction(spec.phenotypes.var1)
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno,
            predictors=[spec.SNPs, spec.phenotypes.var1, inter],
            test="linear",
        )

        # Gathering the observed matrix
        matrix = modelspec.create_data_matrix(self.phenotypes, self.genotypes)

        # Checking we caught the GWAS interaction
        self.assertTrue(modelspec.has_gwas_interaction,
                        "The modelspec did not catch the GWAS interaction")

        # Checking the shape of the matrix
        self.assertEqual((self.data.shape[0], 3), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[self.data.index, outcome_col]
        self.assertTrue(outcomes.equals(self.data.pheno),
                        "The outcomes are not as expected")

        # Checking the predictor
        np.testing.assert_array_equal(
            matrix.loc[self.data.index, spec.phenotypes.var1.id].values,
            self.data.var1.values,
            err_msg="The predictor 'var1' is not as expected.",
        )

        # Checking the resulting column for the GWAS interaction
        multiplication_dict = modelspec.gwas_interaction
        self.assertEqual(1, len(multiplication_dict),
                         "Wrong number of interaction multiplication")
        self.assertTrue(inter.id in multiplication_dict,
                        "Wrong result column for the GWAS interaction")

        # Checking the columns to multiply for the GWAS interaction
        multiplication_cols = multiplication_dict[inter.id]
        self.assertEqual(1, len(multiplication_cols),
                         "Wrong number of columns to multiply for the GWAS "
                         "interaction")
        self.assertEqual(spec.phenotypes.var1.id, multiplication_cols[0],
                         "Wrong column to multiply for the GWAS interaction")

    def test_gwas_interaction_complex_category(self):
        """Test a simple GWAS interaction with a categorical value."""
        # Creating the model specification
        var3 = spec.factor(spec.phenotypes.var3)
        var5 = spec.factor(spec.phenotypes.var5)
        inter = spec.gwas_interaction(spec.phenotypes.var1, var3, var5)
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno,
            predictors=[spec.SNPs, spec.phenotypes.var1, var3, var5, inter],
            test="linear",
        )

        # Gathering the observed matrix
        matrix = modelspec.create_data_matrix(self.phenotypes, self.genotypes)

        # Checking we caught the GWAS interaction
        self.assertTrue(modelspec.has_gwas_interaction,
                        "The modelspec did not catch the GWAS interaction")

        # Checking the shape of the matrix
        self.assertEqual((self.data.shape[0], 8), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[self.data.index, outcome_col]
        self.assertTrue(outcomes.equals(self.data.pheno),
                        "The outcomes are not as expected")

        # Checking the predictor (var1)
        np.testing.assert_array_equal(
            matrix.loc[self.data.index, spec.phenotypes.var1.id].values,
            self.data.var1.values,
            err_msg="The predictor 'var1' is not as expected.",
        )

        # Checking the predictor (var3, all level)
        for level in ("x1", "x2"):
            matrix_col = "{}:level.{}".format(var3.id, level)
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, matrix_col].values,
                (self.data.var3 == level).astype(float).values,
                err_msg="The predictor 'var3' (level '{}') is not as "
                        "expected".format(level),
            )

        # Checking the predictor (var5, all level)
        for level in (1, 2, 3):
            matrix_col = "{}:level.{}".format(var5.id, level)
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, matrix_col].values,
                (self.data.var5 == level).astype(float).values,
                err_msg="The predictor 'var5' (level '{}') is not as "
                        "expected".format(level),
            )

        # Checking the resulting column for the GWAS interaction
        multiplication_dict = modelspec.gwas_interaction
        self.assertEqual(6, len(multiplication_dict),
                         "Wrong number of interaction multiplication")

        # Checking the columns to multiply
        for var3_l in ("x1", "x2"):
            for var5_l in (1, 2, 3):
                result_col = "{}:level.{}:level.{}".format(
                    inter.id, var3_l, var5_l,
                )
                self.assertTrue(
                    result_col in multiplication_dict,
                    "Wrong result column for the GWAS interaction with "
                    "var3 level '{}' and var5 level "
                    "'{}'".format(var3_l, var5_l),
                )

                multiplication_cols = multiplication_dict[result_col]
                self.assertEqual(
                    3, len(multiplication_cols),
                    "Wrong number of columns to multiply for the GWAS "
                    "interaction with level var3 level '{}' and var5 level "
                    "'{}'".format(var3_l, var5_l),
                )
                self.assertEqual(
                    {spec.phenotypes.var1.id,
                     "{}:level.{}".format(var3.id, var3_l),
                     "{}:level.{}".format(var5.id, var5_l)},
                    set(multiplication_cols),
                    "Wrong column to multiply for the GWAS interaction with "
                    "var3 level '{}' and var5 level "
                    "'{}'".format(var3_l, var5_l),
                )





# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import unittest

import numpy as np
import pandas as pd
from scipy.stats import binom

from geneparse import parsers

from .. import modelspec as spec
from ..phenotypes.dummy import _DummyPhenotypes
from ..statistics.models.linear import StatsLinear


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


class TestGrammar(unittest.TestCase):
    """Tests the grammar for the ModelSpec class."""
    @classmethod
    def setUpClass(cls):
        # Creating random data
        cls.data = pd.DataFrame(
            dict(
                pheno=np.random.randint(1, 100, 100),
                var1=np.random.randint(1, 100, 100),
                var2=np.random.rand(100),
                var3=["x{}".format(i) for i in np.random.randint(0, 3, 100)],
                var4=["y{}".format(i) for i in np.random.randint(0, 2, 100)],
                var5=np.random.randint(0, 4, 100),
                snp=binom.rvs(2, 0.3, size=100),
            ),
            index=["sample_{}".format(i+1) for i in range(100)],
        )

        # Changing one factor to categorical data
        cls.data.loc[:, "var5"] = cls.data.var5.astype("category")

        # Creating the dummy phenotype container
        phenotypes = ["pheno"] + ["var{}".format(i+1) for i in range(5)]
        cls.phenotypes = _DummyPhenotypes()
        cls.phenotypes.data = cls.data[phenotypes].copy()

        # Creating the dummy genotype container
        map_info = pd.DataFrame(
            {"chrom": ["3"],
             "pos": [1234],
             "a1": ["T"],
             "a2": ["C"]},
            index=["snp"],
        )
        cls.genotypes = parsers["dataframe"](
            dataframe=cls.data[["snp"]].copy(),
            map_info=map_info,
        )

    def setUp(self):
        # Resetting the model specification
        spec._reset()

        # Reordering the columns and the rows of the phenotype data frame
        self.phenotypes.data = self.phenotypes.data.iloc[
            np.random.permutation(self.phenotypes.data.shape[0]),
            np.random.permutation(self.phenotypes.data.shape[1])
        ]

    def test_simple_formula(self):
        """Tests a simple ModelSpec object."""
        # Parsing the formula and removing the conditions
        model = spec.parse_formula("pheno ~ g(snp) + var1 + var2")
        model["test"] = StatsLinear
        del model["conditions"]

        # Creating the model
        modelspec = spec.ModelSpec(**model)

        # Gathering the observed matrix
        matrix = modelspec.create_data_matrix(
            self.phenotypes, self.genotypes,
        )

        # Checking the shape of the matrix
        self.assertEqual((self.data.shape[0], 5), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[self.data.index, outcome_col]
        self.assertTrue(outcomes.equals(self.data.pheno),
                        "The outcomes are not as expected")

        # Checking the predictors
        predictors = [spec.genotypes.snp, spec.phenotypes.var1,
                      spec.phenotypes.var2]
        translations = modelspec.get_translations()
        for predictor in predictors:
            # Getting the name of the predictor
            name = translations[predictor.id]

            # Comparing the values
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, predictor.id].values,
                self.data[name].values,
                err_msg="The predictor '{}' is not as expected".format(name),
            )

    def test_pow(self):
        """Tests a power transformation."""
        # Parsing the formula and removing the conditions
        model = spec.parse_formula("pheno ~ g(snp) + pow(var1, 4)")
        model["test"] = StatsLinear
        del model["conditions"]

        # Creating the model
        modelspec = spec.ModelSpec(**model)

        # Gathering the observed matrix
        matrix = modelspec.create_data_matrix(
            self.phenotypes, self.genotypes,
        )

        # Checking the shape of the matrix
        self.assertEqual((self.data.shape[0], 4), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[self.data.index, outcome_col]
        self.assertTrue(outcomes.equals(self.data.pheno),
                        "The outcomes are not as expected")

        # Checking the predictors
        predictors = [spec.genotypes.snp, spec.pow(spec.phenotypes.var1, 4)]
        for predictor, power, name in zip(predictors, (1, 4), ("snp", "var1")):
            # Comparing the values
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, predictor.id].values,
                self.data[name].values**power,
                err_msg="The predictor '{}**{}' is not as "
                        "expected".format(name, power),
            )

    def test_log(self):
        """Tests log transformations (log10 and ln)."""
        # Parsing the formula and removing the conditions
        model = spec.parse_formula(
            "pheno ~ g(snp) + var1 + log10(var1) + ln(var1)"
        )
        model["test"] = StatsLinear
        del model["conditions"]

        # Creating the model
        modelspec = spec.ModelSpec(**model)

        # Gathering the observed matrix
        matrix = modelspec.create_data_matrix(
            self.phenotypes, self.genotypes,
        )

        # Checking the shape of the matrix
        self.assertEqual((self.data.shape[0], 6), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[self.data.index, outcome_col]
        self.assertTrue(outcomes.equals(self.data.pheno),
                        "The outcomes are not as expected")

        # Checking the predictors
        translations = modelspec.get_translations()
        for predictor in (spec.genotypes.snp, spec.phenotypes.var1):
            # Getting the name of the predictor
            name = translations[predictor.id]

            # Comparing the values
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, predictor.id].values,
                self.data[name].values,
                err_msg="The predictor '{}' is not as expected".format(name),
            )

        # Checking the log transform
        predictors = (spec.log10(spec.phenotypes.var1),
                      spec.ln(spec.phenotypes.var1))
        for predictor, transform, name in zip(predictors, (np.log10, np.log),
                                              ("var1", "var1")):
            # Comparing the values
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, predictor.id].values,
                transform(self.data[name].values),
                err_msg="The predictor '{}({})' is not as "
                        "expected".format(transform.__name__, name),
            )

    def test_encode_factor(self):
        """Tests with factor."""
        # Parsing the formula and removing the conditions
        model = spec.parse_formula(
            "pheno ~ factor(var3) + factor(var4) + factor(var5)"
        )
        model["test"] = StatsLinear
        del model["conditions"]

        # Creating the model
        modelspec = spec.ModelSpec(**model)

        # Gathering the observed matrix
        matrix = modelspec.create_data_matrix(
            self.phenotypes, self.genotypes,
        )

        # Checking the shape of the matrix
        self.assertEqual((self.data.shape[0], 8), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[self.data.index, outcome_col]
        self.assertTrue(outcomes.equals(self.data.pheno),
                        "The outcomes are not as expected")

        # Checking the factors
        var_names = ("var3", "var4", "var5")
        predictor_zipped = zip(
            tuple(spec.factor(spec.phenotypes[p]) for p in var_names),
            var_names,
            (("x1", "x2"), ("y1", ), (1, 2, 3)),
        )
        for predictor, name, levels in predictor_zipped:
            for level in levels:
                # Getting the name of the column containing the level data
                matrix_col = "{}:level.{}".format(predictor.id, level)

                # Comparing the results
                np.testing.assert_array_equal(
                    matrix.loc[self.data.index, matrix_col].values,
                    (self.data[name] == level).astype(float).values,
                    err_msg="The predictor '{}' (level '{}') is not as "
                            "expected".format(name, level),
                )

    def test_simple_interaction(self):
        """Tests simple interaction."""
        # Parsing the formula and removing the conditions
        model = spec.parse_formula(
            "pheno ~ g(snp) + var1 + g(snp)*var1"
        )
        model["test"] = StatsLinear
        del model["conditions"]

        # Creating the model
        modelspec = spec.ModelSpec(**model)

        # Gathering the observed matrix
        matrix = modelspec.create_data_matrix(
            self.phenotypes, self.genotypes,
        )

        # Checking the shape of the matrix
        self.assertEqual((self.data.shape[0], 5), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[self.data.index, outcome_col]
        self.assertTrue(outcomes.equals(self.data.pheno),
                        "The outcomes are not as expected")

        # Checking the predictors
        for predictor in (spec.genotypes.snp, spec.phenotypes.var1):
            # Getting the name of the predictor
            name = modelspec.get_translations()[predictor.id]

            # Comparing the values
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, predictor.id].values,
                self.data[name].values,
                err_msg="The predictor '{}' is not as expected".format(name),
            )

        # Checking the interaction
        interaction = spec.interaction(spec.genotypes.snp,
                                       spec.phenotypes.var1)
        np.testing.assert_array_equal(
            matrix.loc[self.data.index, interaction.id].values,
            (self.data.snp * self.data.var1).values,
            err_msg="The interaction 'snp*var1' is not as expected",
        )

    def test_multiple_interaction(self):
        """Tests simple interaction."""
        # Parsing the formula and removing the conditions
        model = spec.parse_formula(
            "pheno ~ g(snp) + var1 + var2 + g(snp)*var1*var2"
        )
        model["test"] = StatsLinear
        del model["conditions"]

        # Creating the model
        modelspec = spec.ModelSpec(**model)

        # Gathering the observed matrix
        matrix = modelspec.create_data_matrix(
            self.phenotypes, self.genotypes,
        )

        # Checking the shape of the matrix
        self.assertEqual((self.data.shape[0], 6), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[self.data.index, outcome_col]
        self.assertTrue(outcomes.equals(self.data.pheno),
                        "The outcomes are not as expected")

        # Checking the predictors
        predictors = (spec.genotypes.snp, spec.phenotypes.var1,
                      spec.phenotypes.var2)
        for predictor in predictors:
            # Getting the name of the predictor
            name = modelspec.get_translations()[predictor.id]

            # Comparing the values
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, predictor.id].values,
                self.data[name].values,
                err_msg="The predictor '{}' is not as expected".format(name),
            )

        # Checking the interaction
        interaction = spec.interaction(
            spec.genotypes.snp, spec.phenotypes.var1, spec.phenotypes.var2,
        )
        np.testing.assert_array_equal(
            matrix.loc[self.data.index, interaction.id].values,
            (self.data.snp * self.data.var1 * self.data.var2).values,
            err_msg="The interaction 'snp*var1*var2' is not as expected",
        )

    def test_factor_interaction(self):
        """Tests interaction between a term and a factor."""
        # Parsing the formula and removing the conditions
        model = spec.parse_formula(
            "pheno ~ g(snp) + factor(var5) + g(snp)*factor(var5)"
        )
        model["test"] = StatsLinear
        del model["conditions"]

        # Creating the model
        modelspec = spec.ModelSpec(**model)

        # Gathering the observed matrix
        matrix = modelspec.create_data_matrix(
            self.phenotypes, self.genotypes,
        )

        # Checking the shape of the matrix
        self.assertEqual((self.data.shape[0], 9), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[self.data.index, outcome_col]
        self.assertTrue(outcomes.equals(self.data.pheno),
                        "The outcomes are not as expected")

        # Checking the predictor
        np.testing.assert_array_equal(
            matrix.loc[self.data.index, spec.genotypes.snp.id].values,
            self.data.snp.values,
            err_msg="The predictor 'snp' is not as expected",
        )

        # For all level:
        factor = spec.factor(spec.phenotypes.var5)
        interaction = spec.interaction(spec.genotypes.snp, factor)
        for level in (1, 2, 3):
            # Creating the expected level
            expected_factor_values = (self.data.var5 == level).astype(float)

            # Checking the factor
            col = factor.id + ":level.{}".format(level)
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, col].values,
                expected_factor_values.values,
                err_msg="The predictor 'var5' (level '{}') is not as "
                        "expected".format(level),
            )

            # Checking the interaction
            col = interaction.id + ":level.{}".format(level)
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, col].values,
                (expected_factor_values * self.data.snp).values,
                err_msg="The interaction 'snp*var5' (var5 level '{}') is not "
                        "as expected".format(level),
            )

    def test_complex_interaction(self):
        """Tests a complex interaction between terms and factors."""
        # Parsing the formula and removing the conditions
        model = spec.parse_formula(
            "pheno ~ g(snp) + var2 + factor(var3) + factor(var4) + "
            "        factor(var5) + "
            "        g(snp)*var2*factor(var3)*factor(var4)*factor(var5)"
        )
        model["test"] = StatsLinear
        del model["conditions"]

        # Creating the model
        modelspec = spec.ModelSpec(**model)

        # Gathering the observed matrix
        matrix = modelspec.create_data_matrix(
            self.phenotypes, self.genotypes,
        )

        # Checking the shape of the matrix
        self.assertEqual((self.data.shape[0], 16), matrix.shape,
                         "The observed matrix is not of the right shape")

        # Checking the intercept
        self.assertEqual([1], matrix.intercept.unique().tolist(),
                         "The intercept is not as expected")

        # Checking the outcome
        outcome_col = spec.phenotypes.pheno.id
        outcomes = matrix.loc[self.data.index, outcome_col]
        self.assertTrue(outcomes.equals(self.data.pheno),
                        "The outcomes are not as expected")

        # Checking the predictors
        translations = modelspec.get_translations()
        for predictor in (spec.genotypes.snp, spec.phenotypes.var2):
            # Getting the name of the predictor
            name = translations[predictor.id]

            # Comparing the values
            np.testing.assert_array_equal(
                matrix.loc[self.data.index, predictor.id].values,
                self.data[name].values,
                err_msg="The predictor '{}' is not as expected".format(name),
            )

        # Checking the factors
        var_names = ("var3", "var4", "var5")
        predictor_zipped = zip(
            tuple(spec.factor(spec.phenotypes[name]) for name in var_names),
            var_names,
            (("x1", "x2"), ("y1", ), (1, 2, 3)),
        )
        for predictor, name, levels in predictor_zipped:
            for level in levels:
                # Getting the name of the column containing the level data
                matrix_col = "{}:level.{}".format(predictor.id, level)

                # Comparing the factor
                np.testing.assert_array_equal(
                    matrix.loc[self.data.index, matrix_col].values,
                    (self.data[name] == level).astype(float).values,
                    err_msg="The predictor '{}' (level '{}') is not as "
                            "expected".format(name, level),
                )

        # Checking the interaction
        interaction = spec.interaction(
            spec.genotypes.snp, spec.phenotypes.var2,
            spec.factor(spec.phenotypes.var3),
            spec.factor(spec.phenotypes.var4),
            spec.factor(spec.phenotypes.var5),
        )
        for var3_l in ("x1", "x2"):
            for var4_l in ("y1", ):
                for var5_l in (1, 2, 3):
                    # The expected factor values
                    expected_var3 = (self.data.var3 == var3_l).astype(float)
                    expected_var4 = (self.data.var4 == var4_l).astype(float)
                    expected_var5 = (self.data.var5 == var5_l).astype(float)

                    # The expected interaction
                    expected_interaction = (
                        self.data.snp * self.data.var2 * expected_var3 *
                        expected_var4 * expected_var5
                    )

                    # Checking the interaction
                    matrix_col = "{}:level.{}:level.{}:level.{}".format(
                        interaction.id, var3_l, var4_l, var5_l,
                    )
                    np.testing.assert_array_equal(
                        matrix.loc[self.data.index, matrix_col].values,
                        expected_interaction.values,
                        err_msg=(
                            "The interaction 'snp*var2*var3*var4*var5*' "
                            "(var3 level '{}', var4 level '{} and var5 "
                            "level '{}' is not as expected".format(
                                var3_l, var4_l, var5_l,
                            )
                        ),
                    )



"""
"""


# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import logging
import warnings

import numpy as np

import statsmodels.api as sm
from statsmodels.tools.sm_exceptions import ConvergenceWarning

from ..core import StatsModels, StatsError


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


__all__ = ["StatsCoxPH"]


logger = logging.getLogger(__name__)


class StatsCoxPH(StatsModels):
    @staticmethod
    def _prepare_data(y, X):
        """Prepares the data for the CoxPH analysis."""
        # Checking y has tte and event columns
        missing_cols = {"tte", "event"} - set(y.columns)
        if missing_cols:
            raise ValueError(
                "missing column in y: coxph requires 'tte' and 'event'"
            )
        if len(y.columns) > 3 and "strata" not in y.columns:
            extra = set(y.columns) - {"tte", "event"}
            logger.warning("{}: unknown column in y, will be ignored".format(
                ",".join(extra),
            ))

        # Removing the intercept (if present)
        if "intercept" in X.columns:
            X = X.drop("intercept", axis=1)

        strata = None
        if "strata" in y.columns:
            strata = y.strata

        return y.tte, y.event, X, strata

    def fit(self, y, X):
        """Fit the model.

        Args:
            y (pandas.DataFrame): The vector of endogenous variable.
            X (pandas.DataFrame): The matrix of exogenous variables.

        Warning
        =======
            The y dataframe (outcome) should contain two columns: tte and
            event.

        """
        # Retrieving the data
        tte, event, X, strata = self._prepare_data(y, X)

        # Creating the CoxPH model and fitting it
        model = sm.PHReg(tte, X, status=event, ties="efron", strata=strata)
        fitted = None
        try:
            with warnings.catch_warnings():
                warnings.filterwarnings("error", category=ConvergenceWarning)
                fitted = model.fit()

        except np.linalg.linalg.LinAlgError as e:
            raise StatsError(str(e))

        except ConvergenceWarning as e:
            raise StatsError(str(e).replace(" Check mle_retvals", ""))

        # Getting the results
        out = {}
        parameters = model.exog_names

        # Results about the model fit
        out = {
            "MODEL": {
                "log_likelihood": fitted.llf,
                "nobs": X.shape[0],
                "nevents": np.sum(event),
            },
        }

        # Computing the confidence intervals only once
        conf_ints = fitted.conf_int()

        # Getting the values for each parameters
        for i, param in enumerate(parameters):
            # If GWAS, check that inference could be done on the SNP
            if param == "SNPs" and np.isnan(fitted.pvalues[i]):
                raise StatsError("Inference did not converge.")

            out[param] = {
                "coef": fitted.params[i],
                "std_err": fitted.bse[i],
                "hr": np.exp(fitted.params[i]),
                "hr_lower_ci": np.exp(conf_ints[i, 0]),
                "hr_upper_ci": np.exp(conf_ints[i, 1]),
                "z_value": fitted.tvalues[i],
                "p_value": fitted.pvalues[i],
            }

        return out



"""
"""


# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import numpy as np


__all__ = ["get_maf"]


def get_maf(genotypes, minor, major):
    """Computes the minor allele frequency using genotypes.

    Args:
        genotypes (pandas.Series): The genotypes.
        minor (str): The minor allele.
        major (str): The major allele.

    Returns:
        tuple: Returns the MAF, the minor allele, the major allele and a
               boolean telling if the markers were flip or not.

    Note
    ====
        The frequency is computed using the alternative allele (i.e. the ``2``
        genotype). If there are more ``2`` genotypes than ``0`` genotypes, the
        alternative allele frequency will be higher than 0.5.

    Note
    ====
        When computing the alternative allele frequency, the missing genotypes
        are excluded. If there are no genotypes, ``NaN`` is returned.

    """
    maf = np.nanmean(genotypes) / 2

    if maf > 0.5:
        return 1 - maf, major, minor, True

    return maf, minor, major, False


def get_sex_maf(genotypes, sexes, minor, major):
    """Computes the minor allele frequency for sexual chromosomes.

    Args:
        genotypes (pandas.Series): The genotypes.
        sexes (pandas.Series): The sex (0=female, 1=male).
        minor (str): The minor allele.
        major (str): The major allele.

    Returns:
        tuple: Returns the MAF, the minor allele, the major allele and a
               boolean telling if the markers were flip or not.

    Warning
    =======
        The ``sexes`` vector should have the same order as the genotypes one,
        and the males are coded as ``1``, while females are coded as ``0``.
        Missing sexes should be coded as ``NaN``.

    """
    # Keeping only the non-missing data
    not_missing = ~(np.isnan(genotypes) | np.isnan(sexes))
    if (~not_missing).all():
        return np.nan, minor, major, False

    # keeping only non missing values
    genotypes = genotypes[not_missing]
    sexes = sexes[not_missing]

    # The number of males and females
    nb_males = np.sum(sexes)
    nb_females = sexes.shape[0] - nb_males

    # Computing the MAF
    maf = (
        (-0.5 * np.dot(genotypes, sexes) + np.sum(genotypes))
        / (nb_males + (2 * nb_females))
    )

    if maf > 0.5:
        return 1 - maf, major, minor, True

    return maf, minor, major, False



"""
Subscribers used to format the output of analyses.
"""


# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import sys
import json
import logging
import datetime

try:
    from twilio.rest import Client
    TWILIO = True
except ImportError:
    TWILIO = False


from .modelspec import result as analysis_results


__all__ = ["Print", "ResultsMemory", "RowWriter", "GWASWriter"]


logger = logging.getLogger(__name__)


class Subscriber(object):
    """Abstract class for subscribers."""
    def init(self, modelspec):
        """Method that gets called by analysis to initialize or reinitialize
        the Subscriber with a modelspec.

        """
        self.modelspec = modelspec
        self._reset_subset()

    def close(self):
        pass

    def _update_current_subset(self, info):
        """Updates information on which part of the dataset is analyzed.

        This method get called by analysis before data is pushed.

        Subclasses should not have to modify this method.

        """
        self.subset_info = info
        self.subset_info_str = ";".join(
            "{}:{}".format(*_) for _ in sorted(info.items())
        )

    def _reset_subset(self):
        self.subset_info = None

    def _update_gwas_interaction(self, columns):
        """Updates information (whether there is a gwas interaction or not).

        Args:
            columns: The columns for the interaction.

        """
        pass

    def handle(self, results):
        """Handle results from a statistical test."""
        raise NotImplementedError()

    @staticmethod
    def _apply_translation(translation, results):
        out = {}
        for k in results:
            if k in translation:
                out[translation[k]] = results[k]
            else:
                out[k] = results[k]
        return out


class TwilioSubscriber(Subscriber):
    def __init__(self, account_sid, auth_token, to, from_):
        if not TWILIO:
            raise ImportError("Install twilio to use this subscriber.")

        self.client = Client(account_sid, auth_token)
        self.from_ = from_
        self.to = to

        self.results = []

    def handle(self, result):
        self.results.append(Subscriber._apply_translation(
            self.modelspec.get_translations(),
            result
        ))

    def close(self):
        logger.info("Sending text message to {}.".format(self.to))

        message = datetime.datetime.now().strftime(
            "Analysis finished on %B %d at %H:%M.\n"
        )
        message += (
            "Received results from {} statistical tests.".format(
                len(self.results)
            )
        )

        self.client.messages.create(
            to=self.to,
            from_=self.from_,
            body=message
        )


class ResultsMemory(Subscriber):
    def __init__(self):
        self.results = []

    def handle(self, results):
        self.results.append(Subscriber._apply_translation(
            self.modelspec.get_translations(),
            results
        ))

    def _get_gwas_results(self):
        out = {}
        for result in self.results:
            name = result["SNPs"]["name"]
            out[name] = result
        return out


class Print(Subscriber):
    def __init__(self, raw=False):
        self.raw = raw

    def handle(self, results):
        if self.subset_info:
            results["MODEL"]["subset_info"] = self.subset_info

        if self.raw:
            print(json.dumps(results, indent=2))
            return

        print(json.dumps(Subscriber._apply_translation(
            self.modelspec.get_translations(),
            results
        ), indent=2))


class RowWriter(Subscriber):
    def __init__(self, filename=None, columns=None, header=False, sep="\t",
                 append=False):
        # columns needs to be a list of 2-tuples ('col', result object or str).
        self.header = header
        self.sep = sep
        self.filename = filename

        self._set_columns(columns)

        if filename:
            self._f = open(filename, "a" if append else "w")
        else:
            self._f = None

        if self.header:
            self.print_header()

    def _set_columns(self, columns):
        self.columns = columns

    def print_header(self):
        header = self.sep.join([i[0] for i in self.columns])
        if self.filename is not None:
            self._f.write(header + "\n")
        else:
            print(header)

    def close(self):
        if self._f:
            self._f.close()

    def handle(self, results):
        row = []
        for name, result in self.columns:
            if isinstance(result, str):
                row.append(result)
            else:
                row.append(str(result.get(results)))

        row = self.sep.join(row)
        if self.filename is not None:
            self._f.write(row + "\n")
            self._f.flush()
        else:
            print(row)


class GWASWriter(RowWriter):
    def __init__(self, filename, test, sep="\t"):
        # A flag for updated values because of GWAS interaction
        self._inter_already_updated = False

        # Setting the final columns to None
        self._final_columns = None

        # The columns that are in common for each model
        self._common_cols = [
            ("snp", analysis_results["SNPs"]["name"]),
            ("chr", analysis_results["SNPs"]["chrom"]),
            ("pos", analysis_results["SNPs"]["pos"]),
            ("major", analysis_results["SNPs"]["major"]),
            ("minor", analysis_results["SNPs"]["minor"]),
            ("maf", analysis_results["SNPs"]["maf"]),
            ("n", analysis_results["MODEL"]["nobs"]),
            ("ll", analysis_results["MODEL"]["log_likelihood"]),
        ]

        # The columns that are specific for each model and analysis
        self._specific_cols = []
        self._specific_model_cols = []

        # Linear
        if test == "linear":
            self._specific_cols = [
                ("coef", ("SNPs", "coef")),
                ("se", ("SNPs", "std_err")),
                ("lower", ("SNPs", "lower_ci")),
                ("upper", ("SNPs", "upper_ci")),
                ("t", ("SNPs", "t_value")),
                ("p", ("SNPs", "p_value")),
            ]
            self._specific_model_cols.append(
                ("adj_r2", analysis_results["MODEL"]["r_squared_adj"]),
            )

        # Logistic
        elif test == "logistic":
            self._specific_cols = [
                ("coef", ("SNPs", "coef")),
                ("se", ("SNPs", "std_err")),
                ("lower", ("SNPs", "lower_ci")),
                ("upper", ("SNPs", "upper_ci")),
                ("t", ("SNPs", "t_value")),
                ("p", ("SNPs", "p_value")),
            ]
            self._specific_model_cols.append(
                ("n_events", analysis_results["MODEL"]["nevents"]),
            )

        # CoxPH
        elif test == "coxph":
            self._specific_cols = [
                ("coef", ("SNPs", "coef")),
                ("se", ("SNPs", "std_err")),
                ("hr", ("SNPs", "hr")),
                ("hr_lower", ("SNPs", "hr_lower_ci")),
                ("hr_upper", ("SNPs", "hr_upper_ci")),
                ("z", ("SNPs", "z_value")),
                ("p", ("SNPs", "p_value")),
            ]
            self._specific_model_cols.append(
                ("n_events", analysis_results["MODEL"]["nevents"]),
            )

        # MixedLM
        elif test == "mixedlm":
            self._specific_cols = [
                ("coef", ("SNPs", "coef")),
                ("se", ("SNPs", "std_err")),
                ("lower", ("SNPs", "lower_ci")),
                ("upper", ("SNPs", "upper_ci")),
                ("z", ("SNPs", "z_value")),
                ("p", ("SNPs", "p_value")),
            ]
            self._specific_model_cols.append(
                ("n_groups", analysis_results["MODEL"]["n_groups"]),
            )

        else:
            logger.warning("{}: invalid test: only common columns will be "
                           "written to file.")

        # Calling super __init__
        super().__init__(filename=filename, columns=None, header=True,
                         append=False, sep=sep)

    @property
    def columns(self):
        # Generating the columns from the common ones, the ones specific to the
        # model, and the ones for the GWAS results
        if self._final_columns is None:
            specific_cols = [
                (output_name, analysis_results[col_name][param_name])
                for output_name, (col_name, param_name) in self._specific_cols
            ]

            self._final_columns = (
                self._common_cols + self._specific_model_cols + specific_cols
            )

        return self._final_columns

    def _set_columns(self, columns):
        pass

    def _reset_output_header(self):
        """Resets the columns."""
        # We reset the final columns
        logger.debug("Re-writing the output header")
        self._final_columns = None

        # We re-write the header
        self._f.seek(0)
        self.print_header()

    def _update_current_subset(self, info):
        """Updates information on which part of the dataset is analyzed."""
        # First, check if the subset_info is None (meaning its the first time
        # this function is called
        if self.subset_info is None:
            logger.debug("Updating columns for subgroup analysis")
            # We add the column
            self._common_cols.append(
                ("subgroup", analysis_results["MODEL"]["subset_info"])
            )

            # Resetting the columns
            self._reset_output_header()

        # Calling super
        super()._update_current_subset(info=info)

    def _update_gwas_interaction(self, columns):
        # First, check if the results from is the default 'SNPs' (meaning it's
        # the first time this function is called). Otherwise, we're in a
        # subgroup analysis
        if not self._inter_already_updated:
            logger.debug("Updating columns for GWAS interaction")
            self._specific_cols = [
                ("{}:{}".format(inter_col, col_info[0]),
                 (inter_col, col_info[1][1]))
                for inter_col in columns for col_info in self._specific_cols
            ]

            # Resetting the columns
            self._reset_output_header()

            # We don't need to change the header again, even for subgroup
            # analysis
            self._inter_already_updated = True

        # Calling super
        super()._update_gwas_interaction(columns=columns)

    def handle(self, results):
        # Adding the subgroup, if any
        if self.subset_info is not None:
            results["MODEL"]["subset_info"] = self.subset_info_str

        # Calling super
        super().handle(results=results)


def subscriber_error(message, abort=None):
    """Logs the error from the subscriber."""
    logger.critical(
        "A subscriber for this analysis raised an exception. "
        "This is be because an invalid key was accessed from the results of "
        "the statistical test.\n"
        "Unknown field: '{}'".format(message)
    )
    if abort:
        abort.set()
    sys.exit(1)



"""
"""


# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.

import importlib.util
import os
import logging
from collections import namedtuple


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


__all__ = ["StatsModels", "StatsError", "parse_modelspec"]


# Logging
logger = logging.getLogger(__name__)


# The statistical results
_StatsResults = namedtuple(
    "_StatsResults",
    ["chr", "pos", "snp", "major", "minor", "maf", "n", "stats_n", "stats"],
)


class StatsModels(object):
    def fit(self, y, X):
        """Fit the model.

        Args:
            y (pandas.DataFrame): The vector of endogenous variable.
            X (pandas.DataFrame): The matrix of exogenous variables.

        Returns:
            dict: All available statistics for a given test.

        """
        raise NotImplementedError()

    def __repr__(self):
        return self.__class__.__name__


class StatsError(Exception):
    """An Exception raised if there is any statistical problem."""
    def __init__(self, msg):
        self.message = str(msg)

    def __str__(self):
        return self.message


def parse_modelspec(filename):
    """Dynamically import a ModelSpec instance from a Python file.

    The file needs to define a model variable corresponding to the ModelSpec
    instance.

    """
    spec = importlib.util.spec_from_file_location(
        "model",
        os.path.abspath(filename)
    )
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.model



"""
"""


# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


from .core import PhenotypesContainer


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


__all__ = ["DataFrameContainer"]


class DataFrameContainer(PhenotypesContainer):
    def __init__(self, dataframe):
        self._phenotypes = dataframe

        # Checking for repeated measurements
        self._repeated = self._phenotypes.index.duplicated().any()

        # Renaming the index
        self._phenotypes.index.name = "sample_id"

    def __repr__(self):
        return "DataFramePhenotypes({:,d} samples, {:,d} variables)".format(
            self.get_nb_samples(),
            self.get_nb_variables(),
        )

    def close(self):
        pass

    def merge(self, other):
        """Merge this instance with another."""
        self._phenotypes = self._phenotypes.join(other._phenotypes)

    def get_phenotypes(self, li=None):
        """Returns a dataframe of phenotypes.

        Args:
            li (list): A list of phenotypes to extract (default: None means
                       all phenotypes).

        Returns:
            pandas.DataFrame: A dataframe containing the phenotypes (with the
            sample IDs as index).

        """
        if li is None:
            return self._phenotypes

        # Check that all requested variables are available.
        missing_variables = set(li) - set(self._phenotypes.columns)
        if missing_variables:
            raise KeyError(
                "Some of the requested phenotypes are unavailable in the "
                "PhenotypeContainer: {}.".format(missing_variables)
            )

        return self._phenotypes.loc[:, list(li)]

    def get_nb_samples(self):
        """Returns the number of samples.

        Returns:
            int: The number of samples.

        """
        if self._repeated:
            # There are duplicated samples, so we count the number of unique
            # samples
            return self._phenotypes.index.unique().shape[0]
        else:
            return self._phenotypes.shape[0]

    def get_nb_variables(self):
        """Returns the number of variables.

        Returns:
            int: The number of variables.

        """
        return self._phenotypes.shape[1]

    def is_repeated(self):
        """Check if the phenotypes contain repeated measurements.

        Returns:
            bool: ``True`` if the data contains repeated measurements,
            ``False`` otherwise.

        """
        return self._repeated

    def keep_samples(self, keep):
        """Keeps only a subset of samples.

        Args:
            keep (list): The list of samples to keep.

        """
        self._phenotypes = self._phenotypes[self._phenotypes.index.isin(keep)]

    def set_sex_column(self, sex_column):
        """Sets the sex column (required to get sex from DataFrame).

        Args:
            sex_column (str): The name of the sex column.

        """
        self._sex_column = sex_column





# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import os
import unittest
from tempfile import TemporaryDirectory

import numpy as np
import pandas as pd
from pkg_resources import resource_filename

from geneparse import parsers

from ..statistics.core import StatsError

from .. import analysis
from .. import subscribers
from .. import modelspec as spec
from ..phenotypes.dummy import _DummyPhenotypes


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


class TestStatsCoxPH(unittest.TestCase):
    """Tests the 'StatsCoxPH' class."""
    @classmethod
    def setUpClass(cls):
        # Loading the data
        data = pd.read_csv(
            resource_filename(__name__, "data/statistics/coxph.txt.bz2"),
            sep="\t",
            compression="bz2",
        )

        # Creating the index
        data["sample"] = [
            "s{}".format(i+1) for i in range(data.shape[0])
        ]
        data = data.set_index("sample")

        # Creating the dummy phenotype container
        cls.phenotypes = _DummyPhenotypes()
        cls.phenotypes.data = data.drop(
            ["snp{}".format(i+1) for i in range(4)],
            axis=1,
        )

        # Permuting the sample to add a bit of randomness
        new_sample_order = np.random.permutation(np.copy(data.index.values))

        # Creating the genotypes data frame
        genotypes = data.loc[
            new_sample_order,
            [_ for _ in data.columns if _.startswith("snp")],
        ].copy()

        # Creating the mapping information
        map_info = pd.DataFrame(
            {"chrom": [3, 3, 2, 1],
             "pos": [1234, 9618, 1519, 5871],
             "a1": ["T", "C", "G", "G"],
             "a2": ["C", "A", "T", "A"]},
            index=["snp1", "snp2", "snp3", "snp4"],
        )

        # Creating the genotype parser
        cls.genotypes = parsers["dataframe"](
            dataframe=genotypes,
            map_info=map_info,
        )

        # Creating a temporary directory for the analysis
        cls.tmp_dir = TemporaryDirectory(prefix="genetest_test_coxph_")

    def setUp(self):
        # Resetting the model specification
        spec._reset()

        # Reordering the columns and the rows of the phenotype data frame
        self.phenotypes.data = self.phenotypes.data.iloc[
            np.random.permutation(self.phenotypes.data.shape[0]),
            np.random.permutation(self.phenotypes.data.shape[1])
        ]

    @classmethod
    def tearDownClass(cls):
        # Cleaning the temporary directory
        cls.tmp_dir.cleanup()

    def test_coxph_gwas(self):
        """Tests coxph regression for GWAS."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=dict(tte=spec.phenotypes.tte, event=spec.phenotypes.event),
            predictors=[spec.SNPs, spec.phenotypes.age,
                        spec.phenotypes.var1, gender],
            test="coxph",
        )

        # The output prefix
        out_prefix = os.path.join(self.tmp_dir.name, "results")

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber], output_prefix=out_prefix,
        )
        gwas_results = subscriber._get_gwas_results()

        # Checking the number of results (should be 3)
        self.assertEqual(3, len(gwas_results.keys()))

        # Checking the first marker (snp1)
        results = gwas_results["snp1"]
        self.assertEqual("snp1", results["SNPs"]["name"])
        self.assertEqual(31, results["MODEL"]["nobs"])
        self.assertEqual(15, results["MODEL"]["nevents"])
        self.assertEqual("3", results["SNPs"]["chrom"])
        self.assertEqual(1234, results["SNPs"]["pos"])
        self.assertEqual("C", results["SNPs"]["minor"])
        self.assertEqual("T", results["SNPs"]["major"])
        self.assertAlmostEqual(0.48162903225806442, results["SNPs"]["maf"])

        # Checking the results (according to R)
        self.assertAlmostEqual(3.49824157323489571, results["SNPs"]["coef"])
        self.assertAlmostEqual(1.05744964734605418, results["SNPs"]["std_err"])
        self.assertAlmostEqual(33.0572720291935767, results["SNPs"]["hr"])
        self.assertAlmostEqual(4.1606792782002913,
                               results["SNPs"]["hr_lower_ci"])
        self.assertAlmostEqual(262.645390558627525,
                               results["SNPs"]["hr_upper_ci"])
        self.assertAlmostEqual(3.3081873751763462, results["SNPs"]["z_value"])
        self.assertAlmostEqual(-np.log10(0.0009390195967453607),
                               -np.log10(results["SNPs"]["p_value"]))

        # Checking the second marker (snp2)
        results = gwas_results["snp2"]
        self.assertEqual("snp2", results["SNPs"]["name"])
        self.assertEqual(60, results["MODEL"]["nobs"])
        self.assertEqual(28, results["MODEL"]["nevents"])
        self.assertEqual("3", results["SNPs"]["chrom"])
        self.assertEqual(9618, results["SNPs"]["pos"])
        self.assertEqual("C", results["SNPs"]["minor"])
        self.assertEqual("A", results["SNPs"]["major"])
        self.assertAlmostEqual(0.40833333333333333, results["SNPs"]["maf"])

        # Checking the results (according to R)
        self.assertAlmostEqual(1.13120383793364998, results["SNPs"]["coef"])
        self.assertAlmostEqual(0.30473046375826845, results["SNPs"]["std_err"])
        self.assertAlmostEqual(3.0993854129021017, results["SNPs"]["hr"])
        self.assertAlmostEqual(1.7056446002934191,
                               results["SNPs"]["hr_lower_ci"])
        self.assertAlmostEqual(5.631999735500465,
                               results["SNPs"]["hr_upper_ci"])
        self.assertAlmostEqual(3.7121455596608572, results["SNPs"]["z_value"])
        self.assertAlmostEqual(-np.log10(0.000205509736523446),
                               -np.log10(results["SNPs"]["p_value"]))

        # Checking the third marker (snp3)
        results = gwas_results["snp3"]
        self.assertEqual("snp3", results["SNPs"]["name"])
        self.assertEqual(60, results["MODEL"]["nobs"])
        self.assertEqual(28, results["MODEL"]["nevents"])
        self.assertEqual("2", results["SNPs"]["chrom"])
        self.assertEqual(1519, results["SNPs"]["pos"])
        self.assertEqual("G", results["SNPs"]["minor"])
        self.assertEqual("T", results["SNPs"]["major"])
        self.assertAlmostEqual(0.20833333333333334, results["SNPs"]["maf"])

        # Checking the results (according to R)
        self.assertAlmostEqual(-0.0069461317578822582, results["SNPs"]["coef"])
        self.assertAlmostEqual(0.39831755948730113, results["SNPs"]["std_err"])
        self.assertAlmostEqual(0.9930779368551547, results["SNPs"]["hr"])
        self.assertAlmostEqual(0.4549197711311750,
                               results["SNPs"]["hr_lower_ci"])
        self.assertAlmostEqual(2.167863107413991,
                               results["SNPs"]["hr_upper_ci"])
        self.assertAlmostEqual(-0.01743867824160966,
                               results["SNPs"]["z_value"])
        self.assertAlmostEqual(0.9860866530659741, results["SNPs"]["p_value"])

        # There should be a file for the failed snp4
        self.assertTrue(os.path.isfile(out_prefix + "_failed_snps.txt"))
        with open(out_prefix + "_failed_snps.txt") as f:
            self.assertEqual(
                [["snp4", "Singular matrix"]],
                [line.split("\t") for line in f.read().splitlines()],
            )

    @unittest.expectedFailure
    def test_coxph_gwas_inter(self):
        """Tests coxph regression for GWAS inter (failure: singular matrix)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # The interaction term
        inter = spec.interaction(spec.phenotypes.var1)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=dict(tte=spec.phenotypes.tte, event=spec.phenotypes.event),
            predictors=[spec.SNPs, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, inter],
            test="coxph",
        )

        # The output prefix
        out_prefix = os.path.join(self.tmp_dir.name, "results")

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber], output_prefix=out_prefix,
        )
        gwas_results = subscriber._get_gwas_results()

        # Checking the number of results (should be 3)
        self.assertEqual(3, len(gwas_results.keys()))

        # Checking the first marker (snp1)
        results = gwas_results["snp1"]
        self.assertEqual("snp1", results["SNPs"]["name"])
        self.assertEqual(31, results["MODEL"]["nobs"])
        self.assertEqual(15, results["MODEL"]["nevents"])
        self.assertEqual("3", results["SNPs"]["chrom"])
        self.assertEqual(1234, results["SNPs"]["pos"])
        self.assertEqual("C", results["SNPs"]["minor"])
        self.assertEqual("T", results["SNPs"]["major"])
        self.assertAlmostEqual(0.48162903225806442, results["SNPs"]["maf"])

        # Checking the results (according to R)
        self.assertAlmostEqual(-0.04756704444017357, results[inter.id]["coef"],
                               places=4)
        self.assertAlmostEqual(0.06434818489300186,
                               results[inter.id]["std_err"], places=5)
        self.assertAlmostEqual(0.9535465409954826, results[inter.id]["hr"],
                               places=4)
        self.assertAlmostEqual(0.8405598094456039,
                               results[inter.id]["hr_lower_ci"], places=4)
        self.assertAlmostEqual(1.081720771831991,
                               results[inter.id]["hr_upper_ci"], places=4)
        self.assertAlmostEqual(-0.7392134606324645,
                               results[inter.id]["z_value"], places=3)
        self.assertAlmostEqual(0.45977738842863736,
                               results[inter.id]["p_value"], places=3)

        # Checking the second marker (snp2)
        results = gwas_results["snp2"]
        self.assertEqual("snp2", results["SNPs"]["name"])
        self.assertEqual(60, results["MODEL"]["nobs"])
        self.assertEqual(28, results["MODEL"]["nevents"])
        self.assertEqual("3", results["SNPs"]["chrom"])
        self.assertEqual(9618, results["SNPs"]["pos"])
        self.assertEqual("C", results["SNPs"]["minor"])
        self.assertEqual("A", results["SNPs"]["major"])
        self.assertAlmostEqual(0.40833333333333333, results["SNPs"]["maf"])

        # Checking the results (according to R)
        self.assertAlmostEqual(0.01742308104178164, results[inter.id]["coef"])
        self.assertAlmostEqual(0.02219048572389944,
                               results[inter.id]["std_err"])
        self.assertAlmostEqual(1.0175757482739625, results[inter.id]["hr"])
        self.assertAlmostEqual(0.9742674031703326,
                               results[inter.id]["hr_lower_ci"])
        self.assertAlmostEqual(1.062809245291237,
                               results[inter.id]["hr_upper_ci"])
        self.assertAlmostEqual(0.7851599671392845,
                               results[inter.id]["z_value"])
        self.assertAlmostEqual(0.4323597838912484,
                               results[inter.id]["p_value"])

        # Checking the third marker (snp3)
        results = gwas_results["snp3"]
        self.assertEqual("snp3", results["SNPs"]["name"])
        self.assertEqual(60, results["MODEL"]["nobs"])
        self.assertEqual(28, results["MODEL"]["nevents"])
        self.assertEqual("2", results["SNPs"]["chrom"])
        self.assertEqual(1519, results["SNPs"]["pos"])
        self.assertEqual("G", results["SNPs"]["minor"])
        self.assertEqual("T", results["SNPs"]["major"])
        self.assertAlmostEqual(0.20833333333333334, results["SNPs"]["maf"])

        # Checking the results (according to R)
        self.assertAlmostEqual(0.005873556606557682, results[inter.id]["coef"])
        self.assertAlmostEqual(0.03350856673847168,
                               results[inter.id]["std_err"])
        self.assertAlmostEqual(1.0058908397614570, results[inter.id]["hr"])
        self.assertAlmostEqual(0.94195099563823814,
                               results[inter.id]["hr_lower_ci"])
        self.assertAlmostEqual(1.074170934795214,
                               results[inter.id]["hr_upper_ci"])
        self.assertAlmostEqual(0.17528522339972738,
                               results[inter.id]["z_value"])
        self.assertAlmostEqual(0.8608555220369414,
                               results[inter.id]["p_value"])

        # There should be a file for the failed snp4
        self.assertTrue(os.path.isfile(out_prefix + "_failed_snps.txt"))
        with open(out_prefix + "_failed_snps.txt") as f:
            self.assertEqual(
                [["snp4", "Singular matrix"]],
                [line.split("\t") for line in f.read().splitlines()],
            )

    def test_coxph_snp1(self):
        """Tests coxph regression with the first SNP."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=dict(tte=spec.phenotypes.tte, event=spec.phenotypes.event),
            predictors=[spec.genotypes.snp1, spec.phenotypes.age,
                        spec.phenotypes.var1, gender],
            test="coxph",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information (it's a flip)
        self.assertEqual("snp1", results["snp1"]["name"])
        self.assertEqual(31, results["MODEL"]["nobs"])
        self.assertEqual(15, results["MODEL"]["nevents"])
        self.assertEqual("3", results["snp1"]["chrom"])
        self.assertEqual(1234, results["snp1"]["pos"])
        self.assertEqual("C", results["snp1"]["minor"])
        self.assertEqual("T", results["snp1"]["major"])
        self.assertAlmostEqual(0.48162903225806442, results["snp1"]["maf"])

        # Checking the results (according to R)
        self.assertAlmostEqual(3.49824157323489571, results["snp1"]["coef"])
        self.assertAlmostEqual(1.05744964734605418, results["snp1"]["std_err"])
        self.assertAlmostEqual(33.0572720291935767, results["snp1"]["hr"])
        self.assertAlmostEqual(4.1606792782002913,
                               results["snp1"]["hr_lower_ci"])
        self.assertAlmostEqual(262.645390558627525,
                               results["snp1"]["hr_upper_ci"])
        self.assertAlmostEqual(3.3081873751763462, results["snp1"]["z_value"])
        self.assertAlmostEqual(-np.log10(0.0009390195967453607),
                               -np.log10(results["snp1"]["p_value"]))

        # TODO: Check the other predictors

    @unittest.expectedFailure
    def test_coxph_snp1_inter(self):
        """Tests coxph regression with the first SNP (interaction)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # The interaction term
        inter = spec.interaction(spec.phenotypes.var1, spec.genotypes.snp1)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=dict(tte=spec.phenotypes.tte, event=spec.phenotypes.event),
            predictors=[spec.genotypes.snp1, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, inter],
            test="coxph",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information (it's a flip)
        self.assertEqual("snp1", results["snp1"]["name"])
        self.assertEqual(31, results["MODEL"]["nobs"])
        self.assertEqual(15, results["MODEL"]["nevents"])
        self.assertEqual("3", results["snp1"]["chrom"])
        self.assertEqual(1234, results["snp1"]["pos"])
        self.assertEqual("C", results["snp1"]["minor"])
        self.assertEqual("T", results["snp1"]["major"])
        self.assertAlmostEqual(0.48162903225806442, results["snp1"]["maf"])

        # Checking the results (according to R)
        self.assertAlmostEqual(-0.04756704444017357, results[inter.id]["coef"],
                               places=4)
        self.assertAlmostEqual(0.06434818489300186,
                               results[inter.id]["std_err"], places=5)
        self.assertAlmostEqual(0.9535465409954826, results[inter.id]["hr"],
                               places=4)
        self.assertAlmostEqual(0.8405598094456039,
                               results[inter.id]["hr_lower_ci"], places=4)
        self.assertAlmostEqual(1.081720771831991,
                               results[inter.id]["hr_upper_ci"], places=4)
        self.assertAlmostEqual(-0.7392134606324645,
                               results[inter.id]["z_value"], places=3)
        self.assertAlmostEqual(0.45977738842863736,
                               results[inter.id]["p_value"], places=3)

        # TODO: Check the other predictors

    def test_coxph_snp1_inter_categorical(self):
        """Tests coxph first SNP (interaction, category)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # The interaction term
        inter = spec.interaction(gender, spec.genotypes.snp1)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=dict(tte=spec.phenotypes.tte, event=spec.phenotypes.event),
            predictors=[spec.genotypes.snp1, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, inter],
            test="coxph",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information (it's a flip)
        self.assertEqual("snp1", results["snp1"]["name"])
        self.assertEqual(31, results["MODEL"]["nobs"])
        self.assertEqual(15, results["MODEL"]["nevents"])
        self.assertEqual("3", results["snp1"]["chrom"])
        self.assertEqual(1234, results["snp1"]["pos"])
        self.assertEqual("C", results["snp1"]["minor"])
        self.assertEqual("T", results["snp1"]["major"])
        self.assertAlmostEqual(0.48162903225806442, results["snp1"]["maf"])

        # Checking the results (according to R)
        col = inter.id + ":level.2"
        self.assertAlmostEqual(1.34436505435604148, results[col]["coef"])
        self.assertAlmostEqual(1.77319747152157015, results[col]["std_err"])
        self.assertAlmostEqual(3.8357502743054757, results[col]["hr"])
        self.assertAlmostEqual(0.118713989626776345,
                               results[col]["hr_lower_ci"])
        self.assertAlmostEqual(123.936363465590887,
                               results[col]["hr_upper_ci"], places=6)
        self.assertAlmostEqual(0.7581586799819029, results[col]["z_value"])
        self.assertAlmostEqual(0.448355994218403997, results[col]["p_value"])

        # TODO: Check the other predictors

    def test_coxph_snp2(self):
        """Tests coxph regression with the second SNP."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=dict(tte=spec.phenotypes.tte, event=spec.phenotypes.event),
            predictors=[spec.genotypes.snp2, spec.phenotypes.age,
                        spec.phenotypes.var1, gender],
            test="coxph",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp2", results["snp2"]["name"])
        self.assertEqual(60, results["MODEL"]["nobs"])
        self.assertEqual(28, results["MODEL"]["nevents"])
        self.assertEqual("3", results["snp2"]["chrom"])
        self.assertEqual(9618, results["snp2"]["pos"])
        self.assertEqual("C", results["snp2"]["minor"])
        self.assertEqual("A", results["snp2"]["major"])
        self.assertAlmostEqual(0.40833333333333333, results["snp2"]["maf"])

        # Checking the results (according to R)
        self.assertAlmostEqual(1.13120383793364998, results["snp2"]["coef"])
        self.assertAlmostEqual(0.30473046375826845, results["snp2"]["std_err"])
        self.assertAlmostEqual(3.0993854129021017, results["snp2"]["hr"])
        self.assertAlmostEqual(1.7056446002934191,
                               results["snp2"]["hr_lower_ci"])
        self.assertAlmostEqual(5.631999735500465,
                               results["snp2"]["hr_upper_ci"])
        self.assertAlmostEqual(3.7121455596608572, results["snp2"]["z_value"])
        self.assertAlmostEqual(-np.log10(0.000205509736523446),
                               -np.log10(results["snp2"]["p_value"]))

        # TODO: Check the other predictors

    def test_coxph_snp2_inter(self):
        """Tests coxph regression with the second SNP (interaction)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # The interaction term
        inter = spec.interaction(spec.phenotypes.var1, spec.genotypes.snp2)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=dict(tte=spec.phenotypes.tte, event=spec.phenotypes.event),
            predictors=[spec.genotypes.snp2, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, inter],
            test="coxph",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp2", results["snp2"]["name"])
        self.assertEqual(60, results["MODEL"]["nobs"])
        self.assertEqual(28, results["MODEL"]["nevents"])
        self.assertEqual("3", results["snp2"]["chrom"])
        self.assertEqual(9618, results["snp2"]["pos"])
        self.assertEqual("C", results["snp2"]["minor"])
        self.assertEqual("A", results["snp2"]["major"])
        self.assertAlmostEqual(0.40833333333333333, results["snp2"]["maf"])

        # Checking the results (according to R)
        self.assertAlmostEqual(0.01742308104178164, results[inter.id]["coef"])
        self.assertAlmostEqual(0.02219048572389944,
                               results[inter.id]["std_err"])
        self.assertAlmostEqual(1.0175757482739625, results[inter.id]["hr"])
        self.assertAlmostEqual(0.9742674031703326,
                               results[inter.id]["hr_lower_ci"])
        self.assertAlmostEqual(1.062809245291237,
                               results[inter.id]["hr_upper_ci"])
        self.assertAlmostEqual(0.7851599671392845,
                               results[inter.id]["z_value"])
        self.assertAlmostEqual(0.4323597838912484,
                               results[inter.id]["p_value"])

        # TODO: Check the other predictors

    def test_coxph_snp3(self):
        """Tests coxph regression with the third SNP."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=dict(tte=spec.phenotypes.tte, event=spec.phenotypes.event),
            predictors=[spec.genotypes.snp3, spec.phenotypes.age,
                        spec.phenotypes.var1, gender],
            test="coxph",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp3", results["snp3"]["name"])
        self.assertEqual(60, results["MODEL"]["nobs"])
        self.assertEqual(28, results["MODEL"]["nevents"])
        self.assertEqual("2", results["snp3"]["chrom"])
        self.assertEqual(1519, results["snp3"]["pos"])
        self.assertEqual("G", results["snp3"]["minor"])
        self.assertEqual("T", results["snp3"]["major"])
        self.assertAlmostEqual(0.20833333333333334, results["snp3"]["maf"])

        # Checking the results (according to R)
        self.assertAlmostEqual(-0.0069461317578822582, results["snp3"]["coef"])
        self.assertAlmostEqual(0.39831755948730113, results["snp3"]["std_err"])
        self.assertAlmostEqual(0.9930779368551547, results["snp3"]["hr"])
        self.assertAlmostEqual(0.4549197711311750,
                               results["snp3"]["hr_lower_ci"])
        self.assertAlmostEqual(2.167863107413991,
                               results["snp3"]["hr_upper_ci"])
        self.assertAlmostEqual(-0.01743867824160966,
                               results["snp3"]["z_value"])
        self.assertAlmostEqual(0.9860866530659741, results["snp3"]["p_value"])

        # TODO: Check the other predictors

    def test_coxph_snp3_inter(self):
        """Tests coxph regression with the third SNP (interaction)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # The interaction term
        inter = spec.interaction(spec.phenotypes.var1, spec.genotypes.snp3)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=dict(tte=spec.phenotypes.tte, event=spec.phenotypes.event),
            predictors=[spec.genotypes.snp3, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, inter],
            test="coxph",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp3", results["snp3"]["name"])
        self.assertEqual(60, results["MODEL"]["nobs"])
        self.assertEqual(28, results["MODEL"]["nevents"])
        self.assertEqual("2", results["snp3"]["chrom"])
        self.assertEqual(1519, results["snp3"]["pos"])
        self.assertEqual("G", results["snp3"]["minor"])
        self.assertEqual("T", results["snp3"]["major"])
        self.assertAlmostEqual(0.20833333333333334, results["snp3"]["maf"])

        # Checking the results (according to R)
        self.assertAlmostEqual(0.005873556606557682, results[inter.id]["coef"])
        self.assertAlmostEqual(0.03350856673847168,
                               results[inter.id]["std_err"])
        self.assertAlmostEqual(1.0058908397614570, results[inter.id]["hr"])
        self.assertAlmostEqual(0.94195099563823814,
                               results[inter.id]["hr_lower_ci"])
        self.assertAlmostEqual(1.074170934795214,
                               results[inter.id]["hr_upper_ci"])
        self.assertAlmostEqual(0.17528522339972738,
                               results[inter.id]["z_value"])
        self.assertAlmostEqual(0.8608555220369414,
                               results[inter.id]["p_value"])

        # TODO: Check the other predictors

    def test_coxph_snp4(self):
        """Tests coxph regression with the third SNP."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=dict(tte=spec.phenotypes.tte, event=spec.phenotypes.event),
            predictors=[spec.genotypes.snp4, spec.phenotypes.age,
                        spec.phenotypes.var1, gender],
            test="coxph",
        )

        # Performing the analysis and retrieving the results
        with self.assertRaises(StatsError) as cm:
            analysis.execute(
                self.phenotypes, self.genotypes, modelspec,
                subscribers=[subscribers.ResultsMemory()],
            )
        self.assertEqual("Singular matrix", str(cm.exception))

    def test_coxph_snp4_inter(self):
        """Tests coxph regression with the third SNP (interaction)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # The interaction term
        inter = spec.interaction(spec.phenotypes.var1, spec.genotypes.snp4)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=dict(tte=spec.phenotypes.tte, event=spec.phenotypes.event),
            predictors=[spec.genotypes.snp4, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, inter],
            test="coxph",
        )

        # Performing the analysis and retrieving the results
        with self.assertRaises(StatsError) as cm:
            analysis.execute(
                self.phenotypes, self.genotypes, modelspec,
                subscribers=[subscribers.ResultsMemory()],
            )
        self.assertEqual("Singular matrix", str(cm.exception))





# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import os
import unittest
from tempfile import TemporaryDirectory

import numpy as np
import pandas as pd
from pkg_resources import resource_filename

from pyplink import PyPlink
from geneparse import parsers

from .. import modelspec as spec
from ..subscribers import GWASWriter
from ..analysis import execute_formula
from ..phenotypes.dummy import _DummyPhenotypes


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


class TestGWASWriter(unittest.TestCase):
    """Tests the 'GWASWriter' class."""
    @classmethod
    def setUpClass(cls):
        # Loading the data
        data = pd.read_csv(
            resource_filename(__name__, "data/statistics/factors.txt.bz2"),
            sep="\t",
            compression="bz2",
        ).set_index("sample_id")

        # Creating the dummy phenotype container
        cls.phenotypes = _DummyPhenotypes()
        cls.phenotypes.data = data.drop(
            [col for col in data.columns if col.startswith("snp")],
            axis=1,
        )

        # Creating a temporary directory
        cls.tmp_dir = TemporaryDirectory(prefix="genetest_test_linear_")

        # The plink file prefix
        cls.plink_prefix = os.path.join(cls.tmp_dir.name, "input")

        # Permuting the sample to add a bit of randomness
        new_sample_order = np.random.permutation(np.copy(data.index.values))

        # Creating the BED file
        with PyPlink(cls.plink_prefix, "w") as bed:
            for i in range(3):
                snp = "snp{}".format(i + 1)
                bed.write_genotypes(data.loc[new_sample_order, snp])

        # Creating the BIM file
        with open(cls.plink_prefix + ".bim", "w") as bim:
            print(1, "snp1", 0, 1, "B", "A", sep="\t", file=bim)
            print(1, "snp2", 0, 2, "B", "A", sep="\t", file=bim)
            print(1, "snp3", 0, 3, "B", "A", sep="\t", file=bim)

        # Creating the FAM file
        with open(cls.plink_prefix + ".fam", "w") as fam:
            for sample in new_sample_order:
                print(sample, sample, 0, 0, 0, -9, file=fam)

    @classmethod
    def tearDownClass(cls):
        cls.tmp_dir.cleanup()

    def setUp(self):
        # Resetting the model specification
        spec._reset()

        # Creating the genotype parser
        self.genotypes = parsers["plink"](self.plink_prefix)

        # Reordering the columns and the rows of the phenotype data frame
        self.phenotypes.data = self.phenotypes.data.iloc[
            np.random.permutation(self.phenotypes.data.shape[0]),
            np.random.permutation(self.phenotypes.data.shape[1])
        ]

        # Creating a GWAS writer
        self.output_name = os.path.join(self.tmp_dir.name, "output.txt")
        self.gwas_writer = GWASWriter(self.output_name, "linear")

    def tearDown(self):
        self.genotypes.close()
        self.gwas_writer.close()

    def test_normal_gwas(self):
        """Tests a normal GWAS analysis."""
        # Executing the formula
        execute_formula(
            phenotypes=self.phenotypes,
            genotypes=self.genotypes,
            formula="pheno ~ SNPs + var1 + var2 + factor(var3) + "
                    "factor(var4) + factor(var5)",
            test="linear",
            subscribers=[self.gwas_writer],
            output_prefix="",
        )

        # Checking the file exists
        self.assertTrue(os.path.isfile(self.output_name))

        # Reading the data
        df = pd.read_csv(self.output_name, sep="\t")
        df = df.set_index("snp", drop=False).loc[["snp1", "snp2", "snp3"], :]

        # Checking all the columns are there
        self.assertEqual(
            set(df.columns),
            {"snp", "chr", "pos", "major", "minor", "maf", "n", "ll", "coef",
             "se", "lower", "upper", "t", "p", "adj_r2"},
            "Missing columns in the GWAS Writer",
        )

        # Checking the coefficients (this should be enough to make sure we get
        # the right columns)
        np.testing.assert_array_almost_equal(
            df.coef.values,
            np.array([6.48616317675085, 4.29701960321627, -1.44891512389655]),
        )

    def test_categorical_gwas(self):
        """Tests a categorical GWAS analysis."""
        # Executing the formula
        execute_formula(
            phenotypes=self.phenotypes,
            genotypes=self.genotypes,
            formula="pheno | var4 ~ SNPs + var1 + var2 + factor(var3) + "
                    "factor(var5)",
            test="linear",
            subscribers=[self.gwas_writer],
            output_prefix="",
        )

        # Checking the file exists
        self.assertTrue(os.path.isfile(self.output_name))

        # Reading the data
        df = pd.read_csv(self.output_name, sep="\t")
        df = df.set_index("snp", drop=False).loc[["snp1", "snp2", "snp3"], :]

        # Checking all the columns are there
        self.assertEqual(
            set(df.columns),
            {"snp", "chr", "pos", "major", "minor", "maf", "n", "ll", "coef",
             "se", "lower", "upper", "t", "p", "adj_r2", "subgroup"},
            "Missing columns in the GWAS Writer",
        )

        # Checking the coefficients for the fist subgroup analysis (this should
        # be enough to make sure we get the right columns)
        subgroup = df.subgroup == "var4:y0"
        np.testing.assert_array_almost_equal(
            df.loc[subgroup, "coef"].values,
            np.array([-0.6798617801022, 5.9538885434287, -15.1758958532932]),
        )

        # Checking the coefficients for the second subgroup analysis
        subgroup = df.subgroup == "var4:y1"
        np.testing.assert_array_almost_equal(
            df.loc[subgroup, "coef"].values,
            np.array([12.69000471889982, 4.2496912087917, 7.5291802846083]),
        )

    def test_simple_interaction_gwas(self):
        """Tests a simple interaction GWAS analysis."""
        # Executing the formula
        execute_formula(
            phenotypes=self.phenotypes,
            genotypes=self.genotypes,
            formula="pheno ~ SNPs + var1 + var2 + factor(var3) + "
                    "factor(var4) + factor(var5) + SNPs*var1",
            test="linear",
            subscribers=[self.gwas_writer],
            output_prefix="",
        )

        # Checking the file exists
        self.assertTrue(os.path.isfile(self.output_name))

        # Reading the data
        df = pd.read_csv(self.output_name, sep="\t")
        df = df.set_index("snp", drop=False).loc[["snp1", "snp2", "snp3"], :]

        # Checking all the columns are there
        self.assertEqual(
            set(df.columns),
            {"snp", "chr", "pos", "major", "minor", "maf", "n", "ll",
             "TRANSFORM:GWAS_INTER:coef", "TRANSFORM:GWAS_INTER:se",
             "TRANSFORM:GWAS_INTER:lower", "TRANSFORM:GWAS_INTER:upper",
             "TRANSFORM:GWAS_INTER:t", "TRANSFORM:GWAS_INTER:p", "adj_r2"},
            "Missing columns in the GWAS Writer",
        )

        # Checking the coefficients (this should be enough to make sure we get
        # the right columns)
        np.testing.assert_array_almost_equal(
            df["TRANSFORM:GWAS_INTER:coef"].values,
            np.array([0.4073975765317, -0.04253919820585, -0.05814970176832]),
        )

    def test_complex_interaction_gwas(self):
        """Tests a complex interaction GWAS analysis."""
        # Executing the formula
        execute_formula(
            phenotypes=self.phenotypes,
            genotypes=self.genotypes,
            formula="pheno ~ SNPs + var1 + var2 + factor(var3) + "
                    "factor(var4) + factor(var5) + SNPs*factor(var3)",
            test="linear",
            subscribers=[self.gwas_writer],
            output_prefix=self.output_name,
        )

        # Checking the file exists
        self.assertTrue(os.path.isfile(self.output_name))

        # Reading the data
        df = pd.read_csv(self.output_name, sep="\t")
        df = df.set_index("snp", drop=False).loc[["snp1", "snp2", "snp3"], :]

        # Checking all the columns are there
        self.assertEqual(
            set(df.columns),
            {"snp", "chr", "pos", "major", "minor", "maf", "n", "ll",
             "TRANSFORM:GWAS_INTER:level.x1:coef",
             "TRANSFORM:GWAS_INTER:level.x1:se",
             "TRANSFORM:GWAS_INTER:level.x1:lower",
             "TRANSFORM:GWAS_INTER:level.x1:upper",
             "TRANSFORM:GWAS_INTER:level.x1:t",
             "TRANSFORM:GWAS_INTER:level.x1:p",
             "TRANSFORM:GWAS_INTER:level.x2:coef",
             "TRANSFORM:GWAS_INTER:level.x2:se",
             "TRANSFORM:GWAS_INTER:level.x2:lower",
             "TRANSFORM:GWAS_INTER:level.x2:upper",
             "TRANSFORM:GWAS_INTER:level.x2:t",
             "TRANSFORM:GWAS_INTER:level.x2:p", "adj_r2"},
            "Missing columns in the GWAS Writer",
        )

        # Checking the coefficients (this should be enough to make sure we get
        # the right columns)
        #   var3=x1
        np.testing.assert_array_almost_equal(
            df["TRANSFORM:GWAS_INTER:level.x1:coef"].values,
            np.array([6.15574045825434, 2.2178331615269, -1.94057337856961]),
        )

        #   var3=x2
        np.testing.assert_array_almost_equal(
            df["TRANSFORM:GWAS_INTER:level.x2:coef"].values,
            np.array([21.47033225862596, 2.7136560668229, -0.05622018358578]),
        )

    def test_complex_interaction_categorical_gwas(self):
        """Tests a complex interaction GWAS analysis."""
        # Executing the formula
        execute_formula(
            phenotypes=self.phenotypes,
            genotypes=self.genotypes,
            formula="pheno | var4 ~ SNPs + var1 + var2 + factor(var3) + "
                    "factor(var5) + SNPs*factor(var3)",
            test="linear",
            subscribers=[self.gwas_writer],
            output_prefix=self.output_name,
        )

        # Checking the file exists
        self.assertTrue(os.path.isfile(self.output_name))

        # Reading the data
        df = pd.read_csv(self.output_name, sep="\t")
        df = df.set_index("snp", drop=False).loc[["snp1", "snp2", "snp3"], :]

        # Checking all the columns are there
        self.assertEqual(
            set(df.columns),
            {"snp", "chr", "pos", "major", "minor", "maf", "n", "ll",
             "TRANSFORM:GWAS_INTER:level.x1:coef",
             "TRANSFORM:GWAS_INTER:level.x1:se",
             "TRANSFORM:GWAS_INTER:level.x1:lower",
             "TRANSFORM:GWAS_INTER:level.x1:upper",
             "TRANSFORM:GWAS_INTER:level.x1:t",
             "TRANSFORM:GWAS_INTER:level.x1:p",
             "TRANSFORM:GWAS_INTER:level.x2:coef",
             "TRANSFORM:GWAS_INTER:level.x2:se",
             "TRANSFORM:GWAS_INTER:level.x2:lower",
             "TRANSFORM:GWAS_INTER:level.x2:upper",
             "TRANSFORM:GWAS_INTER:level.x2:t",
             "TRANSFORM:GWAS_INTER:level.x2:p", "adj_r2", "subgroup"},
            "Missing columns in the GWAS Writer",
        )

        # Checking the coefficients (this should be enough to make sure we get
        # the right columns)
        #   var3=x1, var4=y0
        subgroup = df.subgroup == "var4:y0"
        np.testing.assert_array_almost_equal(
            df.loc[subgroup, "TRANSFORM:GWAS_INTER:level.x1:coef"].values,
            np.array([7.9838584795821, 0.8787655737690, -5.4257289463845]),
        )

        #   var3=x2, var4=y0
        np.testing.assert_array_almost_equal(
            df.loc[subgroup, "TRANSFORM:GWAS_INTER:level.x2:coef"].values,
            np.array([9.5318845882662, 2.1792184593254, -3.8901765902700]),
        )

        #   var3=x1, var4=y1
        subgroup = df.subgroup == "var4:y1"
        np.testing.assert_array_almost_equal(
            df.loc[subgroup, "TRANSFORM:GWAS_INTER:level.x1:coef"].values,
            np.array([-7.33292032058131, -2.53148534004492, -14.560258817242]),
        )

        #   var3=x2, var4=y1
        np.testing.assert_array_almost_equal(
            df.loc[subgroup, "TRANSFORM:GWAS_INTER:level.x2:coef"].values,
            np.array([9.91504894770588, -1.35580107828104, -3.1717946298856]),
        )



"""
Semantics for the grako parser.
"""

import os
import logging
import warnings

from .core import (SNPs, interaction, phenotypes, genotypes, factor, pow,
                   log10, ln, gwas_interaction)


logger = logging.getLogger(__name__)


try:
    from .parser import ModelSpecParser
    PARSER_AVAIL = True
except ImportError as e:
    PARSER_AVAIL = False


class ModelSpecSemantics(object):
    def __init__(self):
        self.entities = {}

    def SNPs(self, ast):
        return SNPs

    def labelled_outcome_group(self, ast):
        return {tree["key"]: tree["name"] for tree in ast["tags"]}

    def interaction(self, ast):
        if SNPs in ast["interaction"]:
            # This is a GWAS interaction
            return gwas_interaction(
                *tuple(term for term in ast["interaction"] if term != SNPs),
                name=ast["as_"] if ast["as_"] else "GWAS_INTER"
            )

        else:
            # This is a normal interaction term
            return interaction(*ast["interaction"], name=ast["as_"])

    def phenotype(self, ast):
        return phenotypes[ast["name"]]

    def string(self, ast):
        return ast["str"][1:-1]

    def integer(self, ast):
        return int(ast["int"])

    def genotype(self, ast):
        return genotypes[ast["variant"]]

    def factor(self, ast):
        return factor(ast["phen"], name=ast["as_"])

    def pow(self, ast):
        return pow(ast["phen"], ast["power"], name=ast["as_"])

    def ln(self, ast):
        return ln(ast["phen"])

    def log10(self, ast):
        return log10(ast["phen"])

    def _default(self, ast):
        return ast


def parse_modelspec(s):
    """Use the modelspec grammar to parse kwargs for modelspec."""
    warnings.warn("use 'parse_formula' instead", DeprecationWarning)
    return parse_formula(s)


def parse_formula(f):
    """Use the modelspec grammar to parse a formula for the modelspec."""
    if not PARSER_AVAIL:
        raise RuntimeError("Impossible to parse the grammar. Is the 'grako' "
                           "module installed?")

    parser = ModelSpecParser(parseinfo=False)

    try:
        return parser.parse(
            f, rule_name="model", semantics=ModelSpecSemantics(),
        )

    except:
        msg = ("Something went wrong while parsing the grammar. This might be "
               "because you are using a different 'grako' version than the "
               "one used to compile the grammar. Please perform the following "
               "command:\n\n{command}")

        # Getting the path of the files
        basedir = os.path.abspath(os.path.dirname(__file__))

        # The command to build the parser
        command = [
            "python", "-m", "grako", os.path.join(basedir, "modelspec.ebnf"),
            "-o", os.path.join(basedir, "parser.py"),
        ]

        raise RuntimeError(msg.format(command=" ".join(command)))
